# -*- coding: utf-8 -*-

from collections import defaultdict

try:
    import numpy as np
except ImportError:
    np = None
import pytest

from ..util.testing import requires
from ..units import (
    amount,
    allclose,
    concatenate,
    concentration,
    fold_constants,
    energy,
    get_derived_unit,
    is_unitless,
    linspace,
    logspace_from_lin,
    SI_base_registry,
    unitless_in_registry,
    format_string,
    get_physical_dimensionality,
    to_unitless,
    length,
    magnitude,
    mass,
    time,
    default_unit_in_registry,
    Backend,
    latex_of_unit,
    unit_of,
    unit_registry_to_human_readable,
    units_library,
    volume,
    simplified,
    uniform,
    unit_registry_from_human_readable,
    _sum,
    UncertainQuantity,
    compare_equality,
    default_units as u,
    patched_numpy as pnp,
    default_constants as dc,
)


def test_dimensionality():
    assert mass + 2 * length - 2 * time == energy
    assert amount - 3 * length == concentration
    assert 3 * length == volume


@requires(units_library)
def test_default_units():
    u.metre
    u.second
    u.hour
    u.decimetre
    u.mole
    u.kilogram
    u.ampere
    u.kelvin
    u.candela
    u.molar
    u.per100eV
    u.joule
    u.gray
    u.eV
    u.MeV
    u.metre
    u.decimetre
    u.centimetre
    u.micrometre
    u.nanometre
    u.gram
    u.molar
    u.hour
    u.perMolar_perSecond
    u.per100eV
    u.umol
    u.umol_per_J


@requires(units_library)
def test_allclose():
    assert allclose(42, 42)
    assert allclose(42 * u.meter, 0.042 * u.km)
    assert not allclose(42, 43)
    assert not allclose(42, 42 * u.meter)
    assert not allclose(42, 43 * u.meter)
    assert not allclose(42 * u.meter, 42)

    a = np.linspace(2, 3) * u.second
    b = np.linspace(2 / 3600.0, 3 / 3600.0) * u.hour
    assert allclose(a, b)
    assert allclose(
        [3600 * u.second, 2 * u.metre / u.hour],
        [1 * u.hour, 2 / 3600 * u.metre / u.second],
    )
    c1 = [[3000, 4000], [3000, 4000]] * u.mol / u.metre ** 3
    c2 = [[3000, 4000], [436.2, 5281.89]] * u.mol / u.metre ** 3
    assert not allclose(c1, c2)
    assert allclose(0 * u.second, 0 * u.second)

    # Possibly allow comparison with scalars in future (broadcasting):
    # assert allclose(2, [2, 2])
    # assert allclose([2, 2], 2)

    # assert not allclose(2, [2, 3])
    # assert not allclose([2, 3], 2)

    # assert allclose(2*u.second, [2, 2]*u.second)
    # assert allclose([2, 2]*u.second, 2*u.second)

    # assert not allclose(2*u.second, [2, 3]*u.second)
    # assert not allclose([2, 3]*u.second, 2*u.second)


@requires(units_library)
def test_is_unitless():
    assert not is_unitless(1 * u.second)
    assert is_unitless(1)
    assert is_unitless({"a": 1, "b": 2.0})
    assert not is_unitless({"a": 2, "b": 5.0 * u.second, "c": 3})
    assert is_unitless(7 * u.molar / u.mole * u.dm3)
    assert is_unitless([2, 3, 4])
    assert not is_unitless([2 * u.m, 3 * u.m])
    assert not is_unitless([3, 4 * u.m])
    assert is_unitless(u.dimensionless)  # this was causing RecursionError


@requires(units_library)
def test_unit_of():
    assert compare_equality(unit_of(0.1 * u.metre / u.second), u.metre / u.second)
    assert not compare_equality(
        unit_of(0.1 * u.metre / u.second), u.kilometre / u.second
    )
    assert compare_equality(unit_of(7), 1)
    assert unit_of(u.gray).dimensionality == u.gray.dimensionality
    ref = (u.joule / u.kg).simplified.dimensionality
    assert unit_of(u.gray, simplified=True).dimensionality == ref

    assert compare_equality(unit_of(dict(foo=3 * u.molar, bar=2 * u.molar)), u.molar)
    assert not compare_equality(
        unit_of(dict(foo=3 * u.molar, bar=2 * u.molar)), u.second
    )
    with pytest.raises(Exception):
        unit_of(dict(foo=3 * u.molar, bar=2 * u.second))
    assert not compare_equality(
        unit_of(dict(foo=3 * u.molar, bar=2 * u.molar)), u.mol / u.metre ** 3
    )


@requires(units_library)
def test_to_unitless():
    dm = u.decimetre
    vals = [1.0 * dm, 2.0 * dm]
    result = to_unitless(vals, u.metre)
    assert result[0] == 0.1
    assert result[1] == 0.2
    with pytest.raises(ValueError):
        to_unitless([42, 43], u.metre)

    with pytest.raises(ValueError):
        to_unitless(np.array([42, 43]), u.metre)

    vals = [1.0, 2.0] * dm
    result = to_unitless(vals, u.metre)
    assert result[0] == 0.1
    assert result[1] == 0.2

    length_unit = 1000 * u.metre
    result = to_unitless(1.0 * u.metre, length_unit)
    assert abs(result - 1e-3) < 1e-12

    amount_unit = 1e-9  # nano
    assert abs(to_unitless(1.0, amount_unit) - 1e9) < 1e-6
    assert (
        abs(
            to_unitless(3 / (u.second * u.molar), u.metre ** 3 / u.mole / u.second)
            - 3e-3
        )
        < 1e-12
    )
    assert abs(to_unitless(2 * u.dm3, u.cm3) - 2000) < 1e-12
    assert abs(to_unitless(2 * u.m3, u.dm3) - 2000) < 1e-12
    assert (float(to_unitless(UncertainQuantity(2, u.dm3, 0.3), u.cm3)) - 2000) < 1e-12

    g1 = UncertainQuantity(4.46, u.per100eV, 0)
    g_unit = get_derived_unit(SI_base_registry, "radiolytic_yield")
    assert abs(to_unitless(g1, g_unit) - 4.46 * 1.036e-7) < 1e-9
    g2 = UncertainQuantity(-4.46, u.per100eV, 0)
    assert abs(to_unitless(-g2, g_unit) - 4.46 * 1.036e-7) < 1e-9

    vals = np.array([1.0 * dm, 2.0 * dm], dtype=object)
    result = to_unitless(vals, u.metre)
    assert result[0] == 0.1
    assert result[1] == 0.2

    one_billionth_molar_in_nanomolar = to_unitless(1e-9 * u.molar, u.nanomolar)
    assert one_billionth_molar_in_nanomolar == 1


@pytest.mark.xfail  # TODO: fails in github-actions, passes locally (on 2 different machines..)
@requires(units_library)
def test_UncertainQuantity():
    a = UncertainQuantity([1, 2], u.m, [0.1, 0.2])
    assert a[1] == [2.0] * u.m
    assert (-a)[0] == [-1.0] * u.m
    assert (-a).uncertainty[0] == [0.1] * u.m
    assert (-a)[0] == (a * -1)[0]
    assert (-a).uncertainty[0] == (a * -1).uncertainty[0]
    assert allclose(a, [1, 2] * u.m)


@requires(units_library, "sympy")
def test_to_unitless__sympy():
    import sympy as sp
    for pi in [np.array(sp.pi, dtype=object), sp.pi]:
        assert sp.cos(to_unitless(pi)) == -1
    with pytest.raises(ValueError):
        to_unitless(sp.pi, u.second)


@requires(units_library)
def test_linspace():
    ls = linspace(2 * u.second, 3 * u.second)
    assert abs(to_unitless(ls[0], u.hour) - 2 / 3600.0) < 1e-15


@requires(units_library)
def test_logspace_from_lin():
    ls = logspace_from_lin(2 * u.second, 3 * u.second)
    assert abs(to_unitless(ls[0], u.hour) - 2 / 3600.0) < 1e-15
    assert abs(to_unitless(ls[-1], u.hour) - 3 / 3600.0) < 1e-15


@requires(units_library)
def test_get_derived_unit():
    registry = SI_base_registry.copy()
    registry["length"] = 1e-1 * registry["length"]
    conc_unit = get_derived_unit(registry, "concentration")
    dm = u.decimetre
    assert abs(conc_unit - 1 * u.mole / (dm ** 3)) < 1e-12 * u.mole / (dm ** 3)

    registry = defaultdict(lambda: 1)
    registry["amount"] = 1e-9  # nano
    assert (
        abs(to_unitless(1.0, get_derived_unit(registry, "concentration")) - 1e9) < 1e-6
    )


@requires(units_library)
def test_unit_registry_to_human_readable():
    # Not as much human readable as JSON serializable...
    d = defaultdict(lambda: 1)
    assert unit_registry_to_human_readable(d) == dict(
        (x, (1, 1)) for x in SI_base_registry.keys()
    )

    ur = {
        "length": 1e3 * u.metre,
        "mass": 1e-2 * u.kilogram,
        "time": 1e4 * u.second,
        "current": 1e-1 * u.ampere,
        "temperature": 1e1 * u.kelvin,
        "luminous_intensity": 1e-3 * u.candela,
        "amount": 1e4 * u.mole,
    }
    assert unit_registry_to_human_readable(ur) == {
        "length": (1e3, "m"),
        "mass": (1e-2, "kg"),
        "time": (1e4, "s"),
        "current": (1e-1, "A"),
        "temperature": (1e1, "K"),
        "luminous_intensity": (1e-3, "cd"),
        "amount": (1e4, "mol"),
    }
    assert unit_registry_to_human_readable(ur) != {
        "length": (1e2, "m"),
        "mass": (1e-2, "kg"),
        "time": (1e4, "s"),
        "current": (1e-1, "A"),
        "temperature": (1e1, "K"),
        "luminous_intensity": (1e-3, "cd"),
        "amount": (1e4, "mol"),
    }


@requires(units_library)
def test_unit_registry_from_human_readable():
    hr = unit_registry_to_human_readable(defaultdict(lambda: 1))
    assert hr == dict((x, (1, 1)) for x in SI_base_registry.keys())
    ur = unit_registry_from_human_readable(hr)
    assert ur == dict((x, 1) for x in SI_base_registry.keys())

    hr = unit_registry_to_human_readable(SI_base_registry)
    assert hr == {
        "length": (1.0, "m"),
        "mass": (1.0, "kg"),
        "time": (1.0, "s"),
        "current": (1.0, "A"),
        "temperature": (1.0, "K"),
        "luminous_intensity": (1.0, "cd"),
        "amount": (1.0, "mol"),
    }
    ur = unit_registry_from_human_readable(hr)
    assert ur == SI_base_registry

    ur = unit_registry_from_human_readable(
        {
            "length": (1.0, "m"),
            "mass": (1.0, "kg"),
            "time": (1.0, "s"),
            "current": (1.0, "A"),
            "temperature": (1.0, "K"),
            "luminous_intensity": (1.0, "cd"),
            "amount": (1.0, "mol"),
        }
    )
    assert ur == {
        "length": u.metre,
        "mass": u.kilogram,
        "time": u.second,
        "current": u.ampere,
        "temperature": u.kelvin,
        "luminous_intensity": u.candela,
        "amount": u.mole,
    }

    ur = unit_registry_from_human_readable(
        {
            "length": (1e3, "m"),
            "mass": (1e-2, "kg"),
            "time": (1e4, "s"),
            "current": (1e-1, "A"),
            "temperature": (1e1, "K"),
            "luminous_intensity": (1e-3, "cd"),
            "amount": (1e4, "mol"),
        }
    )
    assert ur == {
        "length": 1e3 * u.metre,
        "mass": 1e-2 * u.kilogram,
        "time": 1e4 * u.second,
        "current": 1e-1 * u.ampere,
        "temperature": 1e1 * u.kelvin,
        "luminous_intensity": 1e-3 * u.candela,
        "amount": 1e4 * u.mole,
    }

    assert ur != {
        "length": 1e2 * u.metre,
        "mass": 1e-3 * u.kilogram,
        "time": 1e2 * u.second,
        "current": 1e-2 * u.ampere,
        "temperature": 1e0 * u.kelvin,
        "luminous_intensity": 1e-2 * u.candela,
        "amount": 1e3 * u.mole,
    }


@requires(units_library)
def test_unitless_in_registry():
    mag = magnitude(unitless_in_registry(3 * u.per100eV, SI_base_registry))
    ref = 3 * 1.0364268834527753e-07
    assert abs(mag - ref) < 1e-14
    ul = unitless_in_registry([3 * u.per100eV, 5 * u.mol / u.J], SI_base_registry)
    assert allclose(ul, [ref, 5], rtol=1e-6)


@requires(units_library)
def test_compare_equality():
    assert compare_equality(3 * u.m, 3 * u.m)
    assert compare_equality(3 * u.m, 3e-3 * u.km)
    assert compare_equality(3e3 * u.mm, 3 * u.m)
    assert not compare_equality(3 * u.m, 2 * u.m)
    assert not compare_equality(3 * u.m, 3 * u.s)
    assert not compare_equality(3 * u.m, 3 * u.m ** 2)
    assert not compare_equality(3 * u.m, np.array(3))
    assert not compare_equality(np.array(3), 3 * u.m)
    assert compare_equality([3, None], [3, None])
    assert not compare_equality([3, None, 3], [3, None, None])
    assert not compare_equality([None, None, 3], [None, None, 2])
    assert compare_equality([3 * u.m, None], [3, None])
    assert not compare_equality([3 * u.m, None], [3 * u.km, None])


@requires(units_library)
def test_get_physical_dimensionality():
    assert get_physical_dimensionality(3 * u.mole) == {"amount": 1}
    assert get_physical_dimensionality([3 * u.mole]) == {"amount": 1}
    assert get_physical_dimensionality(42) == {}


@requires(units_library)
def test_default_unit_in_registry():
    mol_per_m3 = default_unit_in_registry(3 * u.molar, SI_base_registry)
    assert magnitude(mol_per_m3) == 1
    assert mol_per_m3 == u.mole / u.metre ** 3

    assert default_unit_in_registry(3, SI_base_registry) == 1
    assert default_unit_in_registry(3.0, SI_base_registry) == 1


@requires(units_library)
def test__sum():
    # sum() does not work here...
    assert (_sum([0.1 * u.metre, 1 * u.decimetre]) - 2 * u.decimetre) / u.metre == 0


@requires(units_library)
def test_Backend():
    b = Backend()
    with pytest.raises(ValueError):
        b.exp(-3 * u.metre)
    assert abs(b.exp(1234 * u.metre / u.kilometre) - b.exp(1.234)) < 1e-14


@requires(units_library, "numpy")
def test_Backend__numpy():
    import numpy as np

    b = Backend(np)
    b.sum([1000 * u.metre / u.kilometre, 1], axis=0) == 2.0

    with pytest.raises(AttributeError):
        b.Piecewise


@requires("sympy")
def test_Backend__sympy():
    b = Backend("sympy")
    b.sin(b.pi) == 0

    with pytest.raises(AttributeError):
        b.min


@requires(units_library)
def test_format_string():
    assert format_string(3 * u.gram / u.metre ** 2) == ("3", "g/m**2")
    assert format_string(3 * u.gram / u.metre ** 2, tex=True) == (
        "3",
        r"\mathrm{\frac{g}{m^{2}}}",
    )


@requires(units_library)
def test_joule_html():
    joule_htm = "kg&sdot;m<sup>2</sup>/s<sup>2</sup>"
    joule = u.J.dimensionality.simplified
    assert joule.html == joule_htm


@requires(units_library)
def test_latex_of_unit():
    assert latex_of_unit(u.gram / u.metre ** 2) == r"\mathrm{\frac{g}{m^{2}}}"


@requires(units_library)
def test_concatenate():
    a = [1, 2] * u.metre
    b = [2, 3] * u.mm
    ref = [1, 2, 2e-3, 3e-3] * u.metre
    assert allclose(concatenate((a, b)), ref)


@requires(units_library)
def test_pow0():
    a = [1, 2] * u.metre
    b = a ** 0
    assert allclose(b, [1, 1])

    c = a ** 2
    assert allclose(c, [1, 4] * u.m ** 2)


@requires(units_library)
def test_patched_numpy():
    # see https://github.com/python-quantities/python-quantities/issues/152
    assert allclose(pnp.exp(3 * u.joule / (2 * u.cal)), 1.43119335, rtol=1e-5)
    for arg in ([1, 2], [[1], [2]], [1], 2):
        assert np.all(pnp.exp(arg) == np.exp(arg))


@requires(units_library)
def test_tile():
    a = [2 * u.m, 3 * u.km]
    assert allclose(pnp.tile(a, 2), [2 * u.m, 3000 * u.m, 2e-3 * u.km, 3 * u.km])


@requires(units_library)
def test_simplified():
    assert allclose(
        simplified(dc.molar_gas_constant), 8.314 * u.J / u.mol / u.K, rtol=2e-3
    )
    assert simplified(2.0) == 2.0


@requires(units_library)
def test_polyfit_polyval():
    p1 = pnp.polyfit([0, 1, 2], [0, 1, 4], 2)
    assert allclose(p1, [1, 0, 0], atol=1e-14)
    assert allclose(pnp.polyval(p1, 3), 9)
    assert allclose(pnp.polyval(p1, [4, 5]), [16, 25])

    p2 = pnp.polyfit([0, 1, 2] * u.s, [0, 1, 4] * u.m, 2)
    for _p, _r, _a in zip(
        p2,
        [1 * u.m / u.s ** 2, 0 * u.m / u.s, 0 * u.m],
        [0 * u.m / u.s ** 2, 1e-15 * u.m / u.s, 1e-15 * u.m],
    ):
        assert allclose(_p, _r, atol=_a)
    assert allclose(pnp.polyval(p2, 3 * u.s), 9 * u.m)
    assert allclose(pnp.polyval(p2, [4, 5] * u.s), [16, 25] * u.m)


@requires(units_library)
def test_uniform():
    base = [3 * u.km, 200 * u.m]
    refs = [np.array([3000, 200]), np.array([3, 0.2])]

    def _check(case, ref):
        assert np.any(np.all(magnitude(uniform(case)) == ref, axis=1))

    _check(base, refs)
    _check(tuple(base), refs)
    keys = "foo bar".split()
    assert magnitude(uniform(dict(zip(keys, base)))) in [
        dict(zip(keys, r)) for r in refs
    ]


@requires(units_library)
def test_fold_constants():
    assert abs(fold_constants(dc.pi) - np.pi) < 1e-15


@requires("numpy")
def test_to_unitless___0D_array_with_object():
    from ..util._expr import Constant

    # b = Backend('sympy')
    # pi = np.array(b.pi)
    pi = np.array(Constant(np.pi))
    one_thousand = to_unitless(pi * u.metre, u.millimeter)
    assert get_physical_dimensionality(one_thousand) == {}
    assert abs(magnitude(one_thousand) - np.arctan(1) * 4e3) < 1e-12



import warnings
from chempy.util.testing import requires
from chempy.units import units_library
from ..water_diffusivity_holz_2000 import water_self_diffusion_coefficient as w_sd


def test_water_self_diffusion_coefficient():
    warnings.filterwarnings("error")
    assert abs(w_sd(273.15 + 0.0) - 1.099e-9) < 0.027e-9
    assert abs(w_sd(273.15 + 4.0) - 1.261e-9) < 0.011e-9
    assert abs(w_sd(273.15 + 10) - 1.525e-9) < 0.007e-9
    assert abs(w_sd(273.15 + 15) - 1.765e-9) < 0.006e-9
    assert abs(w_sd(273.15 + 20) - 2.023e-9) < 0.001e-9
    assert abs(w_sd(273.15 + 25) - 2.299e-9) < 0.001e-9
    assert abs(w_sd(273.15 + 30) - 2.594e-9) < 0.001e-9
    assert abs(w_sd(273.15 + 35) - 2.907e-9) < 0.004e-9

    try:
        w_sd(1)
    except UserWarning:
        pass  # good warning raised
    else:
        raise
    warnings.resetwarnings()


@requires(units_library)
def test_water_self_diffusion_coefficient__units():
    from chempy.units import allclose, linspace, default_units as u

    unit = u.m ** 2 / u.s
    assert allclose(
        1e9 * w_sd(298.15 * u.K, units=u), 2.299 * unit, rtol=1e-3, atol=1e-8 * unit
    )
    assert allclose(
        1e9 * w_sd(linspace(297, 299) * u.K, units=u),
        2.299 * u.m ** 2 / u.s,
        rtol=5e-2,
        atol=1e-2 * unit,
    )



import warnings

from chempy.units import allclose

from ..water_permittivity_bradley_pitzer_1979 import water_permittivity
from chempy.util.testing import requires
from chempy.units import linspace, units_library, default_units as u


def test_water_permittivity():
    warnings.filterwarnings("error")
    abs(water_permittivity(273.15 + 0) - 80) < 1.0
    abs(water_permittivity(273.15 + 20) - 80.1) < 0.2
    abs(water_permittivity(273.15 + 100) - 55.3) < 0.5

    try:
        water_permittivity(1)
    except UserWarning:
        pass  # good: warning raised
    else:
        raise
    warnings.resetwarnings()


@requires(units_library)
def test_water_permittivity__units():
    assert allclose(
        water_permittivity(298.15 * u.K, 1 * u.bar, units=u), 78.38436874203077
    )
    assert allclose(
        water_permittivity(linspace(297.5, 298.65) * u.K, 1 * u.bar, units=u),
        78,
        rtol=1e-2,
        atol=1e-2,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io
from itertools import chain
import os
import re
import shutil
import subprocess
import sys
import warnings

from setuptools import setup

pkg_name = "chempy"
url = "https://github.com/bjodah/" + pkg_name
license = "BSD-2-Clause"

RELEASE_VERSION = os.environ.get("%s_RELEASE_VERSION" % pkg_name.upper(), "")  # v*


def _path_under_setup(*args):
    return os.path.join(os.path.dirname(__file__), *args)


release_py_path = _path_under_setup(pkg_name, "_release.py")

if len(RELEASE_VERSION) > 0:
    if RELEASE_VERSION[0] == "v":
        TAGGED_RELEASE = True
        __version__ = RELEASE_VERSION[1:]
    else:
        raise ValueError("Ill formatted version")
else:
    TAGGED_RELEASE = False
    # read __version__ attribute from _release.py:
    exec(open(release_py_path).read())
    if __version__.endswith("git"):
        try:
            _git_version = (
                subprocess.check_output(["git", "describe", "--dirty"])
                .rstrip()
                .decode("utf-8")
                .replace("-dirty", ".dirty")
            )
        except subprocess.CalledProcessError:
            warnings.warn(
                "A git-archive is being installed - version information incomplete."
            )
        else:
            if "develop" not in sys.argv:
                warnings.warn("Using git to derive version: dev-branches may compete.")
                _ver_tmplt = (
                    r"\1.post\2"
                    if os.environ.get("CONDA_BUILD", "0") == "1"
                    else r"\1.post\2+\3"
                )
                __version__ = re.sub(
                    r"v([0-9.]+)-(\d+)-(\S+)", _ver_tmplt, _git_version
                )  # .dev < '' < .post

submodules = [
    "chempy.electrochemistry",
    "chempy.kinetics",
    "chempy.printing",
    "chempy.properties",
    "chempy.thermodynamics",
    "chempy.util",
]

tests = [
    "chempy.tests",
    "chempy.electrochemistry.tests",
    "chempy.kinetics.tests",
    "chempy.printing.tests",
    "chempy.properties.tests",
    "chempy.thermodynamics.tests",
    "chempy.util.tests",
]

classifiers = [
    "Development Status :: 4 - Beta",
    "Operating System :: OS Independent",
    "Topic :: Scientific/Engineering",
    "Topic :: Scientific/Engineering :: Chemistry",
    "Programming Language :: Python",
]

with io.open(_path_under_setup(pkg_name, "__init__.py"), "rt", encoding="utf-8") as f:
    short_description = f.read().split('"""')[1].strip()
if not 10 < len(short_description) < 255:
    warnings.warn("Short description from __init__.py probably not read correctly")
long_descr = io.open(_path_under_setup("README.rst"), encoding="utf-8").read()
if not len(long_descr) > 100:
    warnings.warn("Long description from README.rst probably not read correctly.")
_author, _author_email = open(_path_under_setup("AUTHORS"), "rt").readline().split("<")

extras_req = {
    "integrators": ["pyodeint>=0.10.4", "pycvodes>=0.14.5", "pygslodeiv2>=0.9.4"],
    "solvers": ["pykinsol>=0.1.6"],
    "native": ["pycompilation>=0.4.12", "pycodeexport>=0.1.3", "appdirs"],
    "docs": ["Sphinx", "sphinx_rtd_theme", "numpydoc"],
    "plotting": ["bokeh>=0.13.0", "ipywidgets"],
    "testing": ["black", "pytest>=3.9", "pytest-cov", "pytest-flake8", "rstcheck"],
}
extras_req["all"] = list(chain(extras_req.values()))

setup_kwargs = dict(
    name=pkg_name,
    version=__version__,
    description=short_description,
    long_description=long_descr,
    author=_author.strip(),
    author_email=_author_email.split(">")[0].strip(),
    url=url,
    license=license,
    keywords=["chemistry", "water properties", "physical chemistry"],
    packages=[pkg_name] + submodules + tests,
    classifiers=classifiers,
    install_requires=[
        "numpy>1.11.3",
        "scipy>=1.0.1",
        "matplotlib>=2.2.3",
        "sympy>=1.1.1,!=1.2",
        "quantities>=0.12.1",
        "pyneqsys>=0.5.7",
        "pyodesys>=0.14.5" if sys.version_info[0] >= 3 else "pyodesys<0.12",
        "pyparsing>=2.0.3",
        "sym>=0.3.7",
        "tabulate",
        "pulp>=1.6.8",
        "dot2tex>=2.11.3",
    ],
    extras_require=extras_req,
    python_requires=">=3.8",
)

if __name__ == "__main__":
    try:
        if TAGGED_RELEASE:
            # Same commit should generate different sdist
            # depending on tagged version (set CHEMPY_RELEASE_VERSION)
            # this will ensure source distributions contain the correct version
            shutil.move(release_py_path, release_py_path + "__temp__")
            open(release_py_path, "wt").write(
                "__version__ = '{}'\n".format(__version__)
            )
        setup(**setup_kwargs)
    finally:
        if TAGGED_RELEASE:
            shutil.move(release_py_path + "__temp__", release_py_path)



# -*- coding: utf-8 -*-

from collections import OrderedDict

from chempy import Substance
from chempy.util.testing import requires
from chempy.util.parsing import parsing_library
from ..table import as_per_substance_html_table
from ..web import html


@requires(parsing_library)
def test_as_per_substance_html_table():
    substances = OrderedDict([(k, Substance.from_formula(k)) for k in "H OH".split()])
    assert html(as_per_substance_html_table([2, 3], substances)).count("<tr>") == 3
    assert (
        html(as_per_substance_html_table({"H": 2, "OH": 3}, substances)).count("<tr>")
        == 3
    )



# -*- coding: utf-8 -*-

import shutil
import subprocess
import tempfile

import pytest


from chempy.util.table import rsys2tablines, rsys2table, rsys2pdf_table
from .test_graph import _get_rsys
from ..testing import skipif


try:
    pdflatex_missing = subprocess.call(["pdflatex", "--version"]) != 0
except OSError:
    pdflatex_missing = True


def test_rsys2tablines():
    assert rsys2tablines(_get_rsys(), tex=False) == ["1 & 2 A & -> & B & 3 & - & None"]


def test_rsys2table():
    assert rsys2table(_get_rsys()) == (
        r"""
\begin{table}
\centering
\label{tab:none}
\caption[None]{None}
\begin{tabular}{lllllll}
\toprule
Id. & Reactants &  & Products & {Rate constant} & Unit & Ref \\
\midrule
1 & \ensuremath{2 \boldsymbol{A}} & \ensuremath{\rightarrow} &"""
        + r""" \ensuremath{\boldsymbol{B}} & \ensuremath{3} & \ensuremath{-} & None \\
\bottomrule
\end{tabular}
\end{table}"""
    )


@pytest.mark.parametrize("longtable", (True, False))
@skipif(pdflatex_missing, reason="latex not installed? (pdflatex command missing)")
def test_rsys2pdf_table(longtable):
    rsys = _get_rsys()
    tempdir = tempfile.mkdtemp()
    try:
        rsys2pdf_table(rsys, tempdir, longtable=longtable)
    finally:
        shutil.rmtree(tempdir)


@pytest.mark.skipif(
    pdflatex_missing, reason="latex not installed? (pdflatex command missing)"
)
def test_rsys2pdf_table_no_output_dir():
    rsys = _get_rsys()
    rsys2pdf_table(rsys, save=False)



# -*- coding: utf-8 -*-

from itertools import product
import math

try:
    import numpy as np
except ImportError:
    np = None

from .printing import number_to_scientific_html
from ._util import get_backend, mat_dot_vec, prodpow


class EqCalcResult(object):

    attrs = {
        "sane": bool,
        "success": bool,
        "nfev": int,
        "njev": int,
        "time_cpu": float,
        "time_wall": float,
    }

    def __init__(self, eqsys, init_concs, varied):
        self.eqsys = eqsys
        self.all_inits, self.varied_keys = self.eqsys.per_substance_varied(
            init_concs, varied
        )
        self.conc = np.empty_like(self.all_inits)
        for k, v in self.attrs.items():
            setattr(self, k, np.zeros(self.all_inits.shape[:-1], dtype=v))

    def solve(self, **kwargs):
        for index in product(*map(range, self.all_inits.shape[:-1])):
            slc = tuple(index) + (slice(None),)
            self.conc[slc], nfo, sane = self.eqsys._solve(self.all_inits[slc], **kwargs)
            self.sane[index] = sane

            def _get(k):
                try:
                    return nfo[k]
                except TypeError:
                    return nfo[-1][k]

            for k in self.attrs:
                if k == "sane":
                    continue
                try:
                    getattr(self, k)[index] = _get(k)
                except KeyError:
                    pass

    def _repr_html_(self):
        def fmt(num):
            return number_to_scientific_html(num, fmt=5)

        if len(self.varied_keys) == 0:
            raise NotImplementedError()
        elif len(self.varied_keys) == 1:
            var_html = self.eqsys.substances[self.varied_keys[0]].html_name
            header = ["[%s]<sub>0</sub>" % var_html] + [
                "[%s]" % s.html_name for s in self.eqsys.substances.values()
            ]

            def row(i):
                j = self.eqsys.as_substance_index(self.varied_keys[0])
                return map(fmt, [self.all_inits[i, j]] + self.conc[i, :].tolist())

            pre = "  <td style='font-weight: bold;'>\n      "
            linker = "\n    </td>\n    <td>\n      "
            post = "\n    </td>"
            rows = [
                pre + linker.join(row(i)) + post for i in range(self.all_inits.shape[0])
            ]
            template = """<table>\n  <tr>\n    <th>\n    %s\n    </th>\n  </tr>\n  <tr>\n  %s\n  </tr>\n</table>"""
            head_linker = "\n    </th>\n    <th>\n      "
            row_linker = "\n  </tr>\n  <tr>\n  "
            return template % (head_linker.join(header), row_linker.join(rows))
        else:
            raise NotImplementedError()

    def plot(
        self,
        ls=("-", "--", ":", "-."),
        c=("k", "r", "g", "b", "c", "m", "y"),
        latex=None,
    ):
        import matplotlib.pyplot as plt

        if latex is None:
            latex = next(iter(self.eqsys.substances.values())).latex_name is not None
        if len(self.varied_keys) == 0:
            raise NotImplementedError()
        elif len(self.varied_keys) == 1:
            x = self.all_inits[:, self.eqsys.as_substance_index(self.varied_keys[0])]
            for idx, (k, v) in enumerate(self.eqsys.substances.items()):
                lbl = (r"$\mathrm{" + v.latex_name + "}$") if latex else v.name
                plt.plot(
                    x,
                    self.conc[:, idx],
                    label=lbl,
                    ls=ls[idx % len(ls)],
                    c=c[idx % len(c)],
                )

            ax = plt.gca()

            # Log-log
            ax.set_xscale("log")
            ax.set_yscale("log")

            # Axis labels
            var_latex = self.eqsys.substances[self.varied_keys[0]].latex_name
            ax.set_xlabel((r"$[\mathrm{%s}]_0$" if latex else "[%s]0") % var_latex)
            ax.set_ylabel("Concentration")

            # Outside legend
            box = ax.get_position()
            ax.set_position([box.x0, box.y0, box.width * 0.75, box.height])
            # Put a legend to the right of the current axis
            ax.legend(loc="upper left", bbox_to_anchor=(1, 1))
        else:
            raise NotImplementedError()


class _NumSys(object):

    small = 0  # precipitation limit
    pre_processor = None
    post_processor = None
    internal_x0_cb = None

    def __init__(
        self,
        eqsys,
        rref_equil=False,
        rref_preserv=False,
        backend=None,
        precipitates=(),
        new_eq_params=True,
    ):
        self.eqsys = eqsys
        self.rref_equil = rref_equil
        self.rref_preserv = rref_preserv
        self.backend = get_backend(backend)
        self.precipitates = precipitates
        self.new_eq_params = new_eq_params

    def _get_A_ks(self, eq_params):
        non_precip_rids = self.eqsys.non_precip_rids(self.precipitates)
        return self.eqsys.stoichs_constants(
            self.eqsys.eq_constants(non_precip_rids, eq_params, self.small),
            self.rref_equil,
            backend=self.backend,
            non_precip_rids=non_precip_rids,
        )

    def _inits_and_eq_params(self, params):
        eq_params = params[self.eqsys.ns :]
        if not self.new_eq_params:
            assert not eq_params, "Adjust number of parameters accordingly"
            eq_params = None  # use those of eqsys
        return params[: self.eqsys.ns], eq_params


class NumSysLin(_NumSys):
    def internal_x0_cb(self, init_concs, params):
        # reduce risk of stationary starting point
        return (99 * init_concs + self.eqsys.dissolved(init_concs)) / 100

    def f(self, yvec, params):
        from pyneqsys.symbolic import linear_exprs

        init_concs, eq_params = self._inits_and_eq_params(params)
        A, ks = self._get_A_ks(eq_params)
        # yvec == C
        f_equil = [q / k - 1 if k != 0 else q for q, k in zip(prodpow(yvec, A), ks)]
        B, comp_nrs = self.eqsys.composition_balance_vectors()
        f_preserv = linear_exprs(
            B, yvec, mat_dot_vec(B, init_concs), rref=self.rref_preserv
        )
        return f_equil + f_preserv


class _NumSysLinNegPenalty(NumSysLin):
    def f(self, yvec, params):
        import sympy as sp

        f_penalty = [sp.Piecewise((yi ** 2, yi < 0), (0, True)) for yi in yvec]
        return super(_NumSysLinNegPenalty, self).f(yvec, params) + f_penalty


class NumSysLinRel(NumSysLin):
    def max_concs(self, params, min_=min, dtype=np.float64):
        init_concs = params[: self.eqsys.ns]
        return self.eqsys.upper_conc_bounds(init_concs, min_=min_, dtype=dtype)

    def pre_processor(self, x, params):
        return x / self.max_concs(params), params

    def post_processor(self, x, params):
        return x * self.max_concs(params), params

    def f(self, yvec, params):
        import sympy as sp

        return NumSysLin.f(
            self,
            [
                m * yi
                for m, yi in zip(
                    self.max_concs(params, min_=lambda x: sp.Min(*x), dtype=object),
                    yvec,
                )
            ],
            params,
        )


class NumSysSquare(NumSysLin):

    small = 1e-35

    def pre_processor(self, x, params):
        return (np.sqrt(np.abs(x)), params)

    def post_processor(self, x, params):
        return x ** 2, params

    def internal_x0_cb(self, init_concs, params):
        return np.sqrt(np.abs(init_concs))

    def f(self, yvec, params):
        ysq = [yi * yi for yi in yvec]
        return NumSysLin.f(self, ysq, params)


class NumSysLinTanh(NumSysLin):
    def pre_processor(self, x, params):
        ymax = self.eqsys.upper_conc_bounds(params[: self.eqsys.ns])
        return np.arctanh((8 * x / ymax - 4) / 5), params

    def post_processor(self, x, params):
        ymax = self.eqsys.upper_conc_bounds(params[: self.eqsys.ns])
        return ymax * (4 + 5 * np.tanh(x)) / 8, params

    def internal_x0_cb(self, init_concs, params):
        return self.pre_processor(init_concs, init_concs)[0]

    def f(self, yvec, params):
        import sympy

        ymax = self.eqsys.upper_conc_bounds(
            params[: self.eqsys.ns],
            min_=lambda a, b: sympy.Piecewise((a, a < b), (b, True)),
        )
        ytanh = [yimax * (4 + 5 * sympy.tanh(yi)) / 8 for yimax, yi in zip(ymax, yvec)]
        return NumSysLin.f(self, ytanh, params)


class NumSysLog(_NumSys):

    small = math.exp(-36)  # anything less than `small` is insignificant

    def pre_processor(self, x, params):
        return (
            np.log(np.asarray(x) + NumSysLog.small),  # 10: damping
            params,
        )  # zero conc. ~= small

    def post_processor(self, x, params):
        return np.exp(x), params

    def internal_x0_cb(self, init_concs, params):
        # return [1]*len(init_concs)
        return [0.1] * len(init_concs)

    def f(self, yvec, params):
        from pyneqsys.symbolic import linear_exprs

        init_concs, eq_params = self._inits_and_eq_params(params)
        A, ks = self._get_A_ks(eq_params)
        # yvec == ln(C)
        f_equil = mat_dot_vec(A, yvec, [-self.backend.log(k) for k in ks])
        B, comp_nrs = self.eqsys.composition_balance_vectors()
        f_preserv = linear_exprs(
            B,
            list(map(self.backend.exp, yvec)),
            mat_dot_vec(B, init_concs),
            rref=self.rref_preserv,
        )
        return f_equil + f_preserv



# -*- coding: utf-8 -*-
"""
Non-public API (expect changes without notice).

Helper functions for using native code generation together with pyodesys.
"""

from collections import OrderedDict

try:
    from pyodesys.native import native_sys
except ImportError:
    native_sys = None
    PartiallySolvedSystem = None
    render_mako = None
else:
    from pyodesys.symbolic import PartiallySolvedSystem
    from pyodesys.native.util import render_mako


from .. import Substance

_anon = """
    template <typename T>
    constexpr T vecmin(T&& a){
        return std::forward<T>(a);
    }
    template <typename T1, typename T2>
    constexpr typename std::common_type<T1, T2>::type vecmin(T1&& a, T2&& b){
        return (a < b) ? std::forward<T1>(a) : std::forward<T2>(b);
    }
    template <typename T, typename... Ts>
    constexpr typename std::common_type<T, Ts...>::type vecmin(T&& a, Ts&&... args){
        return vecmin(std::forward<T>(a), vecmin(std::forward<Ts>(args)...));
    }
    std::vector<double> upper_conc_bounds(const double * const y){
        auto bounds = std::vector<double>(${odesys.ny});
        double cc[${ncomp}];
      % for ci in range(ncomp):
        cc[${ci}] = ${' + '.join([('%d*y[%d]' % (v, k)) if v != 1 else 'y[%d]' % k for k, v in comp_conc[ci].items()])};
      % endfor
      % for si, subst_key in enumerate(getattr(odesys, 'free_names', odesys.names)):
       % if len(subst_comp[si]) > 0:
        bounds[${si}] = vecmin(${', '.join(['INFINITY' if n == 0 else ('cc[%d]/%d' % (ci, n)) if n != 1 else 'cc[%d]' % ci for ci, n in subst_comp[si].items()])});
       % else:
        bounds[${si}] = INFINITY;
       % endif
      % endfor
        return bounds;
    }
"""  # noqa


_first_step = """
    m_upper_bounds = upper_conc_bounds(${init_conc});
    m_lower_bounds.resize(${odesys.ny});
    return m_rtol*std::min(get_dx_max(x, y), 1.0);
"""

_roots_ss = """
    const int ny = get_ny();
    std::vector<double> f(ny);
    double tot=0.0;
    rhs(x, y, &f[0]);
    for (int i=0; i<ny; ++i){
        tot += std::min(std::abs(f[i]/m_atol[i]), std::abs(f[i]/y[i]/m_rtol));  // m_atol needs to be of size ny!
    }
    out[0] = tot/ny - m_special_settings[0];
    this->nrev++;
    return AnyODE::Status::success;
"""

_constr_special_settings = r"""
    if (m_special_settings.size() == 0){
         std::cerr << __FILE__ << ":" << __LINE__ << ": no special_settings passed, using default [%(factor)s]\n";
         m_special_settings = {%(factor)s};
    } else {
         // std::cerr << __FILE__ << ":" << __LINE__ << ": using special_settings:" << m_special_settings[0] << "\n";
    }
""" % {
    "factor": "1e2"
}


def _get_comp_conc(rsys, odesys, comp_keys, skip_keys):
    comp_conc = []
    for comp_key in comp_keys:
        if comp_key in skip_keys:
            continue  # see Substance.__doc__
        _d = OrderedDict()
        for si, subst_key in enumerate(odesys.names):
            coeff = rsys.substances[subst_key].composition.get(comp_key, 0)
            if coeff != 0:
                _d[si] = coeff
        comp_conc.append(_d)
    return comp_conc


def _get_subst_comp(rsys, odesys, comp_keys, skip_keys):
    subst_comp = []
    for subst_key in odesys.names:
        _d = OrderedDict()
        for k, v in rsys.substances[subst_key].composition.items():
            if k in skip_keys:
                continue
            _d[comp_keys.index(k)] = v
        subst_comp.append(_d)
    return subst_comp


def get_native(
    rsys, odesys, integrator, skip_keys=(0,), steady_state_root=False, conc_roots=None
):
    comp_keys = Substance.composition_keys(
        rsys.substances.values(), skip_keys=skip_keys
    )
    if PartiallySolvedSystem is None:
        raise ValueError("Failed to import 'native_sys' from 'pyodesys.native'")
    elif isinstance(odesys, PartiallySolvedSystem):
        init_conc = "&m_p[%d]" % (len(odesys.params) - len(odesys.original_dep))
    else:
        init_conc = "y"

    kw = dict(
        namespace_override={
            "p_get_dx_max": True,
        }
    )
    if all(subst.composition is None for subst in rsys.substances.values()):
        pass
    else:
        kw["namespace_override"]["p_anon"] = render_mako(
            _anon,
            odesys=odesys,
            ncomp=len(comp_keys),
            comp_conc=_get_comp_conc(rsys, odesys, comp_keys, skip_keys),
            subst_comp=_get_subst_comp(rsys, odesys, comp_keys, skip_keys),
        )
        kw["namespace_override"]["p_first_step"] = render_mako(
            _first_step, init_conc=init_conc, odesys=odesys
        )
    ns_extend = kw.get("namespace_extend", {})

    if steady_state_root or conc_roots:
        if not native_sys[integrator]._NativeCode._support_roots:
            raise ValueError("integrator '%s' does not support roots." % integrator)
        if odesys.roots is not None:
            raise ValueError("roots already set")
    if steady_state_root:
        assert conc_roots is None
        kw["namespace_override"]["p_nroots"] = " return 1; "
        kw["namespace_override"]["p_roots"] = _roots_ss
        if "p_constructor" not in ns_extend:
            ns_extend["p_constructor"] = []
        ns_extend["p_constructor"] += [_constr_special_settings]
    elif conc_roots:
        # This could (with some effort) be rewritten to take limits as parameters and have a
        # preprocessor in odesys.pre_processors do the dedimensionalization.
        assert not steady_state_root
        assert all(k in odesys.names and k in rsys.substances for k in conc_roots)
        kw["namespace_override"]["p_nroots"] = " return %d; " % len(conc_roots)
        kw["namespace_override"]["p_roots"] = (
            "".join(
                [
                    "    out[%(i)d] = y[%(j)d] - m_special_settings[%(i)d];\n"
                    % dict(i=i, j=odesys.names.index(k))
                    for i, k in enumerate(conc_roots)
                ]
            )
            + "    return AnyODE::Status::success;\n"
        )
        if "p_constructor" not in ns_extend:
            ns_extend["p_constructor"] = []
        ns_extend["p_constructor"] += [
            'if (m_special_settings.size() != %d) throw std::runtime_error("special_settings missing");'
            % len(conc_roots)
        ]

    if "p_includes" not in ns_extend:
        ns_extend["p_includes"] = set()
    ns_extend["p_includes"] |= {"<type_traits>", "<vector>"}
    return native_sys[integrator].from_other(odesys, namespace_extend=ns_extend, **kw)



# https://github.com/alexanderkuk/log-progress
def log_progress(sequence, every=None, size=None, name="Items"):
    from ipywidgets import IntProgress, HTML, VBox
    from IPython.display import display

    is_iterator = False
    if size is None:
        try:
            size = len(sequence)
        except TypeError:
            is_iterator = True
    if size is not None:
        if every is None:
            if size <= 200:
                every = 1
            else:
                every = int(size / 200)  # every 0.5%
    else:
        assert every is not None, "sequence is iterator, set every"

    if is_iterator:
        progress = IntProgress(min=0, max=1, value=1)
        progress.bar_style = "info"
    else:
        progress = IntProgress(min=0, max=size, value=0)
    label = HTML()
    box = VBox(children=[label, progress])
    display(box)

    index = 0
    try:
        for index, record in enumerate(sequence, 1):
            if index == 1 or index % every == 0:
                if is_iterator:
                    label.value = "{name}: {index} / ?".format(name=name, index=index)
                else:
                    progress.value = index
                    label.value = u"{name}: {index} / {size}".format(
                        name=name, index=index, size=size
                    )
            yield record
    except Exception:
        progress.bar_style = "danger"
        raise
    else:
        progress.bar_style = "success"
        progress.value = index
        label.value = "{name}: {index}".format(name=name, index=str(index or "?"))



import warnings
from .._util import _any


A, B, C = 1.1709, 0.001827, 89.93
eta20_cP = 1.0020


def water_viscosity(T=None, eta20=None, units=None, warn=True):
    """Viscosity of water (cP) as function of temperature (K)

    Parameters
    ----------
    T : float
        Temperature (in Kelvin) (default: 298.15 K)
    eta20 : float
        Viscosity of water at 20 degree Celsius.
    units : object (optional)
        object with attributes: kelvin & centipoise
    warn : bool
        Emit UserWarning when outside temperature range.

    Returns
    -------
    Water viscosity at temperature ``T``.

    """
    if units is None:
        cP = 1
        K = 1
    else:
        cP = units.centipoise
        K = units.kelvin
    if T is None:
        T = 298.15 * K
    if eta20 is None:
        eta20 = eta20_cP * cP
    t = T - 273.15 * K
    if warn and (_any(t < 0 * K) or _any(t > 100 * K)):
        warnings.warn("Temperature is outside range (0-100 degC)")
    # equation (5) in the paper says "log" but they seem to mean "log10"
    # when comparing with Table II.
    return eta20 * 10 ** ((A * (20 - t) - B * (t - 20) ** 2) / (t + C))


reference = dict(
    doi="10.1021/j100721a006",
    url="https://doi.org/10.1021/j100721a006",
    year=1969,
    month="jan",
    publisher="American Chemical Society ({ACS})",
    volume=73,
    number=1,
    pages=(34, 39),
    author="Lawrence Korson and Walter Drost-Hansen and Frank J. Millero",
    title="Viscosity of water at various temperatures",
    journal="The Journal of Physical Chemistry",
)



# -*- coding: utf-8 -*-

from collections import OrderedDict
from ..chemistry import Substance
from .numbers import number_to_scientific_html


class Table(object):
    def __init__(self, rows, headers=None):
        self.rows, self.headers = rows, headers

    def _html(self, printer, **kwargs):
        def map_fmt(cont, fmt, joiner="\n"):
            return joiner.join(map(lambda x: fmt % printer._print(x, **kwargs), cont))

        rows = [map_fmt(self.headers, "<th>%s</th>")] + [
            map_fmt(row, "<td>%s</td>") for row in self.rows
        ]
        return "<table>%s</table>" % map_fmt(rows, "<tr>%s</tr>")

    def _repr_html_(self):
        from .web import html

        return html(self)


def as_per_substance_html_table(
    cont, substances=None, header=None, substance_factory=Substance.from_formula
):
    """ """
    if substances is None:
        substances = OrderedDict([(k, substance_factory(k)) for k in cont])

    def _elem(k):
        try:
            return cont[k]
        except (IndexError, TypeError):
            return cont[list(substances.keys()).index(k)]

    rows = [
        (v.html_name, number_to_scientific_html(_elem(k)))
        for k, v in substances.items()
    ]
    return Table(rows, ["Substance", header or ""])



# -*- coding: utf-8 -*-

from .util.pyutil import ChemPyDeprecationWarning

import warnings

from .kinetics.eyring import eyring_equation, EyringParam, EyringParamWithUnits


warnings.warn("use .kinetics.eyring instead of .eyring", ChemPyDeprecationWarning)



# -*- coding: utf-8 -*-

import math
from operator import add
from functools import reduce

import pytest

from chempy import Substance
from chempy.units import (
    allclose,
    units_library,
    default_constants,
    Backend,
    to_unitless,
    SI_base_registry,
    default_units as u,
)
from ..testing import requires
from ..pyutil import defaultkeydict
from .._expr import (
    Expr,
    mk_Poly,
    mk_PiecewisePoly,
    create_Piecewise,
    create_Poly,
    Log10,
    Constant,
    Symbol as _Symbol,
)
from ..parsing import parsing_library


def test_Constant():
    c23 = Constant(2.3)
    assert abs(2.3 - float(c23)) < 1e-15

    c_div = Constant(2.3) / Constant(3.4)
    assert abs(2.3 / 3.4 - float(c_div)) < 1e-15


class HeatCapacity(Expr):
    parameter_keys = ("temperature",)


class EinsteinSolid(HeatCapacity):
    parameter_keys = HeatCapacity.parameter_keys + ("molar_gas_constant",)
    argument_names = ("einstein_temperature", "molar_mass")

    def __call__(self, variables, backend=math):
        TE, molar_mass = self.all_args(
            variables, backend=backend
        )  # einstein_temperature
        T, R = self.all_params(variables, backend=backend)
        # Canonical ensemble:
        molar_c_v = 3 * R * (TE / (2 * T)) ** 2 * backend.sinh(TE / (2 * T)) ** -2
        return molar_c_v / molar_mass


def _get_cv(kelvin=1, gram=1, mol=1):

    Al = Substance.from_formula("Al", data={"DebyeT": 428 * kelvin, "mass": 26.9815385})
    Be = Substance.from_formula("Be", data={"DebyeT": 1440 * kelvin, "mass": 9.0121831})

    def einT(s):
        return 0.806 * s.data["DebyeT"]

    return {s.name: EinsteinSolid([einT(s), s.mass * gram / mol]) for s in (Al, Be)}


@requires(parsing_library)
def test_Expr():
    cv = _get_cv()
    _ref = 0.8108020083055849
    assert (
        abs(cv["Al"]({"temperature": 273.15, "molar_gas_constant": 8.3145}) - _ref)
        < 1e-14
    )
    with pytest.raises(Exception):
        float(cv["Al"])


def _poly(args, x, backend=None):
    x0, coeffs = args[0], args[1:]
    return reduce(add, [c * (x - x0) ** i for i, c in enumerate(coeffs)])


@requires(parsing_library)
def test_Expr__nested_Expr():
    Poly = Expr.from_callback(
        _poly, parameter_keys=("x",), argument_names=("x0", Ellipsis)
    )
    T = Poly([3, 7, 5])

    cv = _get_cv()
    _ref = 0.8108020083055849
    args = {"temperature": T, "x": (273.15 - 7) / 5 + 3, "molar_gas_constant": 8.3145}
    assert abs(cv["Al"](args) - _ref) < 1e-14
    Al2 = cv["Al"] / 2
    assert abs(Al2(args) - _ref / 2) < 1e-14


def test_nargs():
    class A(Expr):
        nargs = 1

    with pytest.raises(ValueError):
        A([1, 2])


@requires("sympy")
def test_Expr_symbolic():
    import sympy

    cv = _get_cv()
    R, T = sympy.symbols("R T")
    sexpr = cv["Be"]({"temperature": T, "molar_gas_constant": R}, backend=sympy)
    assert sexpr.free_symbols == set([T, R])


@requires(units_library)
def test_Expr_units():
    cv = _get_cv(u.kelvin, u.gram, u.mol)
    R = default_constants.molar_gas_constant.rescale(u.joule / u.mol / u.kelvin)

    def _check(T=273.15 * u.kelvin):
        result = cv["Be"](
            {"temperature": T, "molar_gas_constant": R}, backend=Backend()
        )
        ref = 0.7342617587256584 * u.joule / u.gram / u.kelvin
        assert abs(to_unitless((result - ref) / ref)) < 1e-10

    _check()
    _check(491.67 * u.rankine)


@requires(units_library)
def test_Expr_dedimensionalisation__1():
    cv = _get_cv(u.kelvin, u.gram, u.mol)
    units, expr = cv["Be"].dedimensionalisation(SI_base_registry)
    assert units == [u.kelvin, u.kg / u.mol]
    assert abs(expr.args[0] - 0.806 * 1440) < 1e-14
    assert abs(expr.args[1] - 9.01218e-3) < 1e-7


@requires(units_library)
def test_Expr_dedimensionalisation__2():
    Poly = Expr.from_callback(
        _poly, parameter_keys=("E",), argument_names=("x0", Ellipsis)
    )
    T = Poly([3 * u.J, 7 * u.K, 5 * u.K / u.J])
    T = Poly([0.7170172084130019 * u.cal, 12.6 * u.Rankine, 5 * u.K / u.J])

    _ref = 0.8108020083055849  # Al at 273.15 K with R=8.3145
    cv_Al = _get_cv(u.kelvin, u.gram, u.mol)["Al"]

    assert isinstance(cv_Al, EinsteinSolid)
    assert cv_Al.args[0] == 0.806 * 428 * u.kelvin
    assert (
        abs(
            cv_Al(
                {
                    "temperature": 273.15 * u.K,
                    "molar_gas_constant": 8.3145 * u.J / u.K / u.mol,
                }
            )
            - _ref * u.J / u.gram / u.kelvin
        )
        < 1e-14
    )

    cv_Al_units, Al_dedim = cv_Al.dedimensionalisation(SI_base_registry)
    assert allclose(cv_Al_units, [u.K, u.kg / u.mol])
    assert isinstance(Al_dedim, EinsteinSolid)
    T_units, dT = T.dedimensionalisation(SI_base_registry)
    assert allclose(T_units, [u.J, u.K, u.K / u.J])
    assert allclose(dT.args, [3, 7, 5])
    assert (
        abs(
            Al_dedim({"temperature": 273.15, "molar_gas_constant": 8.3145})
            - _ref * 1000
        )
        < 1e-14
    )
    assert (
        abs(
            Al_dedim(
                {
                    "temperature": dT,
                    "E": (273.15 - 7) / 5 + 3,
                    "molar_gas_constant": 8.3145,
                }
            )
            - _ref * 1000
        )
        < 1e-14
    )


@requires(units_library)
def test_Expr_dedimensionalisation__nested():
    Poly = Expr.from_callback(
        _poly, parameter_keys=("E",), argument_names=("x0", Ellipsis)
    )
    TE = Poly([3 * u.J, 7 * u.K, 5 * u.K / u.J])
    TE = (
        Poly([0.7170172084130019 * u.cal, 12.6 * u.Rankine, 5 * u.K / u.J])
        * 0.806
        * 428
        / 273.15
    )

    _ref = 0.8108020083055849  # Al at 273.15 K with R=8.3145
    cv_Al = _get_cv(u.kelvin, u.gram, u.mol)["Al"]
    cv_Al_units, Al_dedim = cv_Al.dedimensionalisation(
        SI_base_registry, {"einstein_temperature": TE}
    )
    assert (
        abs(
            Al_dedim(
                {
                    "temperature": 273.15,
                    "E": (273.15 - 7) / 5 + 3,
                    "molar_gas_constant": 8.3145,
                }
            )
            - _ref * 1000
        )
        < 1e-14
    )


def test_Expr__from_callback():
    def two_dim_gauss(args, x, y, backend=None):
        A, x0, y0, sx, sy = args
        xp, yp = x - x0, y - y0
        vx, vy = 2 * sx ** 2, 2 * sy ** 2
        return A * backend.exp(-(xp ** 2 / vx + yp ** 2 / vy))

    TwoDimGauss = Expr.from_callback(two_dim_gauss, parameter_keys=("x", "y"), nargs=5)
    with pytest.raises(ValueError):
        TwoDimGauss([1, 2])
    args = [3, 2, 1, 4, 5]
    g1 = TwoDimGauss(args)
    ref = two_dim_gauss(args, 6, 7, math)
    assert abs(g1({"x": 6, "y": 7}) - ref) < 1e-14


def test_mk_Poly():
    Poly = mk_Poly("T", reciprocal=True)
    p = Poly([3, 2, 5, 7, 8, 2, 9])
    assert p.eval_poly({"T": 13}) == 2.57829
    assert p.parameter_keys == ("T",)


def test_Expr__nargs():
    class Linear(Expr):
        """Arguments: p0, p1"""

        nargs = 2
        parameter_keys = ("x",)

        def __call__(self, variables, backend=None):
            p0, p1 = self.all_args(variables)
            return p0 + p1 * variables["x"]

    l1 = Linear([3, 2])
    assert l1(dict(x=5)) == 13
    with pytest.raises(ValueError):
        Linear([3])
    with pytest.raises(ValueError):
        Linear([3, 2, 1])

    l2 = Linear([3, 2], ["a", "b"])
    # with pytest.raises(ValueError):
    #     Linear([3, 2], ['a'])
    with pytest.raises(ValueError):
        Linear([3, 2], ["a", "b", "c"])

    assert l2(dict(x=5)) == 13
    assert l2(dict(x=5, a=11, b=13)) == 11 + 13 * 5


def test_PiecewisePoly():
    Poly = mk_Poly("temperature")

    p1 = Poly([0, 1, 0.1])
    assert p1.eval_poly({"temperature": 10}) == 2

    p2 = Poly([0, 3, -0.1])
    assert p2.eval_poly({"temperature": 10}) == 2

    TPiecewisePoly = mk_PiecewisePoly("temperature")
    tpwp = TPiecewisePoly.from_polynomials([(0, 10), (10, 20)], [p1, p2])
    assert tpwp.eval_poly({"temperature": 5}) == 1.5
    assert tpwp.eval_poly({"temperature": 15}) == 1.5
    assert tpwp.parameter_keys == ("temperature",)

    with pytest.raises(ValueError):
        tpwp.eval_poly({"temperature": 21})


def test_create_Piecewise_Poly():
    PolyT = create_Poly("Tmpr")

    p1 = PolyT([1, 0.1])
    assert p1({"Tmpr": 10}) == 2

    p2 = PolyT([3, -0.1])
    assert p2({"Tmpr": 10}) == 2

    PiecewiseT = create_Piecewise("Tmpr")
    pw = PiecewiseT([0, p1, 10, p2, 20])
    assert pw({"Tmpr": 5}) == 1.5
    assert pw({"Tmpr": 15}) == 1.5
    assert pw.parameter_keys == ("Tmpr",)

    with pytest.raises(ValueError):
        pw({"Tmpr": 21})


@requires("sympy")
def test_PiecewisePoly__sympy():
    import sympy as sp

    Poly = mk_Poly("T")
    p1 = Poly([0, 1, 0.1])
    p2 = Poly([0, 3, -0.1])

    TPiecewisePoly = mk_PiecewisePoly("temperature")
    tpwp = TPiecewisePoly([2, 2, 0, 10, 2, 10, 20, 0, 1, 0.1, 0, 3, -0.1])
    x = sp.Symbol("x")
    res = tpwp.eval_poly({"temperature": x}, backend=sp)
    assert isinstance(res, sp.Piecewise)
    assert res.args[0][0] == 1 + 0.1 * x
    assert res.args[0][1] == sp.And(0 <= x, x <= 10)
    assert res.args[1][0] == 3 - 0.1 * x
    assert res.args[1][1] == sp.And(10 <= x, x <= 20)

    with pytest.raises(ValueError):
        tpwp.from_polynomials([(0, 10), (10, 20)], [p1, p2])


@requires("sympy")
def test_create_Piecewise_Poly__sympy():
    import sympy as sp

    Poly = create_Poly("Tmpr")
    p1 = Poly([1, 0.1])
    p2 = Poly([3, -0.1])

    TPw = create_Piecewise("Tmpr")
    pw = TPw([0, p1, 10, p2, 20])
    x = sp.Symbol("x")
    res = pw({"Tmpr": x}, backend=sp)
    ten = sp.S.One * 10
    assert isinstance(res, sp.Piecewise)
    assert res.args[0][0] == 1 + 0.1 * x
    assert res.args[0][1] == sp.And(sp.S.Zero <= x, x <= ten)
    assert res.args[1][0] == 3 - 0.1 * x
    assert res.args[1][1] == sp.And(ten <= x, x <= 2*ten)


@requires("sympy")
def test_create_Piecewise__nan_fallback__sympy():
    import sympy as sp

    TPw = create_Piecewise("Tmpr", nan_fallback=True)
    pw = TPw([0, 42, 10, 43, 20])
    x = sp.Symbol("x")
    res = pw({"Tmpr": x}, backend=sp)
    assert isinstance(res, sp.Piecewise)
    assert res.args[0][0] == 42
    ten = sp.S.One*10
    twenty = 2*ten
    assert res.args[0][1] == sp.And(sp.S.Zero <= x, x <= ten)
    assert res.args[1][0] == 43
    assert res.args[1][1] == sp.And(ten <= x, x <= twenty)
    assert res.args[2][0].name.lower() == "nan"
    assert res.args[2][1] == True  # noqa


def test_BinaryExpr():
    Poly = Expr.from_callback(
        _poly, parameter_keys=("x",), argument_names=("x0", Ellipsis)
    )
    p1 = Poly([1, 2, 3])
    p2 = Poly([2, 3, 4])
    assert p1({"x": 5}) == 14
    assert p2({"x": 5}) == 15
    assert (p1 + p2)({"x": 5}) == 14 + 15
    assert (p1 - p2)({"x": 5}) == 14 - 15
    assert (p1 * p2)({"x": 5}) == 14 * 15
    assert (p1 / p2)({"x": 5}) == 14 / 15

    assert (p1 + 2)({"x": 5}) == 14 + 2
    assert (p1 - 2)({"x": 5}) == 14 - 2
    assert (p1 * 2)({"x": 5}) == 14 * 2
    assert (p1 / 2)({"x": 5}) == 14 / 2

    assert (2 + p1)({"x": 5}) == 2 + 14
    assert (2 - p1)({"x": 5}) == 2 - 14
    assert (2 * p1)({"x": 5}) == 2 * 14
    assert (2 / p1)({"x": 5}) == 2 / 14

    assert p1 + 0 == p1
    assert p1 * 1 == p1
    assert p1 + p2 == p1 + p2
    assert p1 + p2 * 1 == p1 + p2 + 0

    assert -(-p1) == p1


@requires("sympy")
def test_Expr__latex():
    Poly = Expr.from_callback(
        _poly, parameter_keys=("x",), argument_names=("x0", Ellipsis)
    )
    p = Poly([1, 2, 3, 4])
    import sympy

    t = sympy.Symbol("t")
    ref = sympy.latex((2 + 3 * (t - 1) + 4 * (t - 1) ** 2).simplify())
    assert p.latex({"x": "t"}) == ref

    TE = Poly([3, 7, 5])
    cv_Al = _get_cv()["Al"]
    T, E, R, m = sympy.symbols("T E R m")
    _TE = 7 + 5 * (E - 3)
    ref = sympy.latex(
        (
            (3 * R * (_TE / (2 * T)) ** 2 * sympy.sinh(_TE / (2 * T)) ** -2) / m
        ).simplify()
    )
    cv_Al.unique_keys = ("TE_Al", "m_Al")
    res = cv_Al.latex(
        {
            "TE_Al": TE,
            "temperature": "T",
            "x": "E",
            "molar_gas_constant": "R",
            "m_Al": "m",
        }
    )
    assert res == ref

    X = sympy.symbols("X")
    _TE2 = 7 + 5 * (X - 3)
    ref2 = sympy.latex(
        (
            (3 * R * (_TE2 / (2 * T)) ** 2 * sympy.sinh(_TE2 / (2 * T)) ** -2) / m
        ).simplify()
    )
    res2 = cv_Al.latex(
        {"TE_Al": TE, "temperature": "T", "molar_gas_constant": "R", "m_Al": "m"},
        default=lambda k: k.upper(),
    )
    assert res2 == ref2


class Pressure1(Expr):
    argument_names = ("n",)
    parameter_keys = ("temperature", "volume", "R")

    def __call__(self, variables, backend=None):
        (n,) = self.all_args(variables, backend=backend)
        T, V, R = self.all_params(variables, backend=backend)
        return n * R * T / V


def test_Expr__single_arg():
    p = Pressure1(3)
    assert (
        abs(
            p({"temperature": 273.15, "volume": 0.17, "R": 8.314})
            - 3 * 8.314 * 273.15 / 0.17
        )
        < 1e-15
    )


@requires(units_library)
def test_Expr__single_arg__units():
    p = Pressure1(3 * u.mol)
    variables = {
        "temperature": 273.15 * u.kelvin,
        "volume": 170 * u.dm3,
        "R": 8.314 * u.J / u.K / u.mol,
    }
    assert allclose(p(variables), 3 * 8.314 * 273.15 / 0.17 * u.Pa)


class Pressure2(Pressure1):
    def args_dimensionality(self):
        return ({"amount": 1},)


@requires(units_library)
def test_Expr__single_arg__units__dimensionality():
    p = Pressure2(unique_keys=("n1",))
    variables = {
        "temperature": 273.15 * u.kelvin,
        "volume": 170 * u.dm3,
        "R": 8.314 * u.J / u.K / u.mol,
    }
    assert allclose(
        p(dict(n1=3 * u.mol, **variables)), 3 * 8.314 * 273.15 / 0.17 * u.Pa
    )


def test_Expr__argument_defaults():
    class MyExpr(Expr):
        argument_names = ("a", "b", "c")
        argument_defaults = (17, 23)

        def __call__(self, variables={}, backend=math):
            a, b, c = self.all_args(variables, backend=backend)
            return a * b * c

    assert MyExpr([15])() == 15 * 17 * 23
    assert MyExpr([15, 17])() == 15 * 17 * 23
    assert MyExpr([15, 19])() == 15 * 19 * 23
    assert MyExpr([15, 19, 29])() == 15 * 19 * 29
    assert MyExpr(dict(zip("abc", [15, 19, 29])))() == 15 * 19 * 29


class MyK(Expr):
    argument_names = ("H", "S")
    parameter_keys = ("T",)
    R = 8.3145

    def __call__(self, variables, backend=math):
        H, S = self.all_args(variables, backend=backend)
        (T,) = self.all_params(variables, backend=backend)
        return backend.exp(-(H - T * S) / (self.R * T))


def test_Expr__no_args():
    K1 = MyK(unique_keys=("H1", "S1"))
    K2 = MyK(unique_keys=("H2", "S2"))
    add = K1 + K2
    T = 298.15
    res = add({"H1": 2, "H2": 3, "S1": 5, "S2": 7, "T": T})
    RT = 8.3145 * 298.15
    ref = math.exp(-(2 - T * 5) / RT) + math.exp(-(3 - T * 7) / RT)
    assert abs(res - ref) < 1e-14


@requires(units_library)
def test_Expr__equality():
    K1 = MyK(unique_keys=("H1", "S1"))
    K2 = MyK(unique_keys=("H2", "S2"))
    assert K1 != K2
    assert K1 == K1
    K3 = MyK([23e3 * u.J / u.mol, 42 * u.J / u.mol / u.K])
    K4 = MyK([23e3, 42])
    K5 = MyK([24e3 * u.J / u.mol, 42 * u.J / u.mol / u.K])
    K6 = MyK([23e3, 43])
    assert K3 == K3
    assert K4 == K4
    assert K5 != K3
    assert K4 != K6
    assert K3 != K4


@requires("sympy")
def test_Expr__no_args__symbolic():
    K1 = MyK(unique_keys=("H1", "S1"))
    K2 = MyK(unique_keys=("H2", "S2"))
    add = K1 + K2
    import sympy

    v = defaultkeydict(sympy.Symbol)
    res = add(v, backend=sympy)
    R = 8.3145
    expr1 = sympy.exp(-(v["H1"] - v["T"] * v["S1"]) / R / v["T"])
    expr2 = sympy.exp(-(v["H2"] - v["T"] * v["S2"]) / R / v["T"])
    ref = expr1 + expr2
    assert (res - ref).simplify() == 0


class MyK2(Expr):
    argument_names = ("H", "S", "Cp", "Tref")
    argument_defaults = (0, 298.15)
    parameter_keys = "T"
    R = 8.3145

    def __call__(self, variables, backend=math):
        H, S, Cp, Tref = self.all_args(variables, backend=backend)
        (T,) = self.all_params(variables, backend=backend)
        _H = H + Cp * (T - Tref)
        _S = S + Cp * backend.log(T / Tref)
        return backend.exp(-(_H - T * _S) / (self.R * T))


def test_Expr__no_args__arg_defaults():
    K1 = MyK2(unique_keys=("H1", "S1", "Cp1"))
    K2 = MyK2(unique_keys=("H2", "S2"))
    add = K1 + K2
    assert add.all_unique_keys() == set(["H1", "H2", "S1", "S2", "Cp1"])

    T = 293.15
    res = add({"H1": 2, "H2": 3, "S1": 5, "S2": 7, "T": T, "Cp1": 13})
    RT = 8.3145 * T
    H1p = 2 + 13 * (T - 298.15)
    S1p = 5 + 13 * math.log(T / 298.15)
    ref = math.exp(-(H1p - T * S1p) / RT) + math.exp(-(3 - T * 7) / RT)
    assert abs(res - ref) < 1e-14


def test_create_Piecewise():
    PW = create_Piecewise("T")
    Ha, Sa, Hb, Sb, Ta, Tb = 40e3, -60, 37e3, -42, 293.15, 303.15
    a = MyK([Ha, Sa])
    b = MyK([Hb, Sb])
    pw = PW([273.15, a, 298.15, b, 323.15])  # 0, 25, 50 *C
    res_a = pw({"T": Ta})  # 20 *C
    res_b = pw({"T": Tb})  # 30 *C
    ref_a = math.exp(-(Ha - Ta * Sa) / (MyK.R * Ta))
    ref_b = math.exp(-(Hb - Tb * Sb) / (MyK.R * Tb))
    assert abs(res_a - ref_a) < 1e-14
    assert abs(res_b - ref_b) < 1e-14


def test_create_Poly():
    PolyT = create_Poly("T")
    p = PolyT([1, 2, 3, 4, 5])
    assert p({"T": 11}) == 1 + 2 * 11 + 3 * 11 ** 2 + 4 * 11 ** 3 + 5 * 11 ** 4


def test_pow():
    PolyT = create_Poly("T")
    p = PolyT([1, 2, 3])
    p_to_two = p ** 2
    res = p_to_two({"T": 3})
    ref = (1 + 2 * 3 + 3 * 9) ** 2
    assert abs(res - ref) < 1e-12
    two_to_p = 2 ** p
    res = two_to_p({"T": 3})
    ref = 2 ** (1 + 2 * 3 + 3 * 9)
    assert abs(res - ref) < 1e-12


def test_functions():
    PolyT = create_Poly("T")
    p = PolyT([1, 2, 3])
    lgp = Log10(p)
    res = lgp({"T": 3})
    ref = math.log10(1 + 2 * 3 + 3 * 9)
    assert abs(res - ref) < 1e-12


def test_str_arg():
    PolyT = create_Poly("x")
    p = PolyT([1, 2, 3])
    x = Log10("T")
    res = p({"x": x, "T": 1000})
    ref = 1 + 2 * 3 + 3 * 9
    assert abs(res - ref) < 1e-12


def test_implicit_str():
    PolyT = create_Poly("x")
    p = PolyT([1, 2, 3])
    expr1 = p / "u"
    assert abs(expr1({"x": 3, "u": 5}) - (1 + 2 * 3 + 3 * 9) / 5) < 1e-12
    expr2 = "u" / p
    assert abs(expr2({"x": 3, "u": 5}) - 5 / (1 + 2 * 3 + 3 * 9)) < 1e-12
    assert expr1.all_parameter_keys() == set(["x"])
    assert expr2.all_parameter_keys() == set(["x"])


def test_Symbol():
    a = _Symbol.fk("a")
    b = _Symbol.fk("b")
    assert not a == b
    assert not a == b * 0
    assert not a == 0 * b
    assert not a * -0.0 == b * 0.0
    assert not a * 0.0 == b * -0.0



"""
Interactive kinetics app with sliders.
Start by running:
    $ bokeh serve interactive.py
Add --show argument or navigate to:
    http://localhost:5006/interactive
"""
from collections import defaultdict
import sys

from chempy import Reaction, ReactionSystem
from chempy.kinetics.rates import MassAction
from chempy.util.bkh import integration_with_sliders


def get_rsys():
    r1 = Reaction(
        {"A"}, {"B"}, MassAction([4.0 / 100], unique_keys=["k1"]), name="R1: A cons."
    )
    r2 = Reaction(
        {"B", "C"},
        {"A", "C"},
        MassAction([1e4], unique_keys=["k2"]),
        name="R2: A reform.",
    )
    r3 = Reaction(
        {"B": 2}, {"B", "C"}, MassAction([3e7], unique_keys=["k3"]), name="R3: C form."
    )
    return ReactionSystem([r1, r2, r3])


if __name__.startswith("bk_"):
    # TODO: better xlim, ylim, ranges for c0
    from bokeh.io import curdoc

    curdoc().add_root(
        integration_with_sliders(
            get_rsys(),
            tend=1e3,
            c0=defaultdict(float, {"A": 1}),
            parameters={"k1": 4.0 / 100, "k2": 1e4, "k3": 3e7},
            x_axis_type="log",
            y_axis_type="log",
            integrate_kwargs=dict(integrator="cvode"),
        )
    )
elif __name__ == "__main__":
    import warnings

    warnings.warn("Run using 'bokeh serve %s'" % __file__)
    sys.exit(1)



# -*- coding: utf-8 -*-


from .util.pyutil import AttributeContainer


def get_constant_symbols(Symbol=None):
    if Symbol is None:
        from sympy import Symbol
    consts = [
        ("Faraday_constant", "F"),
        ("Avogadro_constant", "N_A"),
        ("vacuum_permittivity", "epsilon_0"),
        ("Boltzmann_constant", "k_B"),
        ("pi", "pi"),
        ("molar_gas_constant", "R"),
    ]
    return AttributeContainer(**{k: Symbol(v) for k, v in consts})



# -*- coding: utf-8 -*-

try:
    import numpy as np
except ImportError:
    np = None
else:
    from ..stoich import (
        get_coeff_mtx,
        decompose_yields,
    )
from chempy.units import units_library, allclose, _sum
from ..testing import requires


@requires("numpy")
def test_get_coeff_mtx():
    r = [({"A": 1}, {"B": 1}), ({"A": 1, "B": 1}, {"C": 2})]
    A = get_coeff_mtx("ABC", r)
    Aref = np.array([[-1, -1], [1, -1], [0, 2]])
    assert np.allclose(A, Aref)


@requires("numpy")
def test_decompose_yields_1():
    from chempy import Reaction

    gamma_yields = {
        "OH-": 0.5,
        "H2O2": 0.7,
        "OH": 2.7,
        "H2": 0.45,
        "H": 0.66,
        "H+": 3.1,
        "HO2": 0.02,
        "e-(aq)": 2.6,
    }

    rxns = [
        Reaction({"H2O": 1}, {"H+": 1, "OH-": 1}),
        Reaction({"H2O": 1}, {"H+": 1, "e-(aq)": 1, "OH": 1}),
        Reaction({"H2O": 1}, {"H": 2, "H2O2": 1}, inact_reac={"H2O": 1}),
        Reaction({"H2O": 1}, {"H2": 1, "H2O2": 1}, inact_reac={"H2O": 1}),
        Reaction({"H2O": 1}, {"H2": 1, "OH": 2}, inact_reac={"H2O": 1}),
        Reaction({"H2O": 1}, {"H2": 3, "HO2": 2}, inact_reac={"H2O": 3}),
    ]

    k = decompose_yields(gamma_yields, rxns)
    k_ref = [0.5, 2.6, 0.33, 0.37, 0.05, 0.01]

    assert np.allclose(k, k_ref)

    G_H2O = sum(rxn.net_stoich(["H2O"])[0] * k[i] for i, rxn in enumerate(rxns))

    assert abs(G_H2O + 4.64) < 1e-3


@requires(units_library)
def test_decompose_yields__units_1():
    from chempy import Reaction
    from chempy.units import default_units as u

    gamma_yields = {
        "OH-": 0.5 * u.per100eV,
        "H2O2": 0.7 * u.per100eV,
        "OH": 2.7 * u.per100eV,
        "H2": 0.45 * u.per100eV,
        "H": 0.66 * u.per100eV,
        "H+": 3.1 * u.per100eV,
        "HO2": 0.02 * u.per100eV,
        "e-(aq)": 2.6 * u.per100eV,
    }

    rxns = [
        Reaction({"H2O": 1}, {"H+": 1, "OH-": 1}),
        Reaction({"H2O": 1}, {"H+": 1, "e-(aq)": 1, "OH": 1}),
        Reaction({"H2O": 1}, {"H": 2, "H2O2": 1}, inact_reac={"H2O": 1}),
        Reaction({"H2O": 1}, {"H2": 1, "H2O2": 1}, inact_reac={"H2O": 1}),
        Reaction({"H2O": 1}, {"H2": 1, "OH": 2}, inact_reac={"H2O": 1}),
        Reaction({"H2O": 1}, {"H2": 3, "HO2": 2}, inact_reac={"H2O": 3}),
    ]

    k = decompose_yields(gamma_yields, rxns)
    k_ref = [0.5, 2.6, 0.33, 0.37, 0.05, 0.01] * u.per100eV

    assert allclose(k, k_ref)

    G_H2O = [rxn.net_stoich(["H2O"])[0] * k[i] for i, rxn in enumerate(rxns)]
    ref = 4.64 * u.per100eV
    assert abs((_sum(G_H2O) + ref) / ref) < 1e-3


@requires("numpy")
def test_decompose_yields_2():
    from chempy import Reaction

    yields = {"B": 3.0, "C": 24.0}
    rxns = [
        Reaction({"A": 1}, {"B": 1, "C": 1}, inact_reac={"A": 1}),
        Reaction({"A": 1}, {"C": 3}),
    ]
    k = decompose_yields(yields, rxns)
    k_ref = [3, 7]

    rtol = 1e-12
    for a, b in zip(k, k_ref):
        assert abs(a - b) < abs(a * rtol)



# -*- coding: utf-8 -*-


from ..arithmeticdict import ArithmeticDict


def test_ArithmeticDict():
    d1 = ArithmeticDict(float, [("a", 1.0), ("b", 2.0)])
    d2 = ArithmeticDict(float, [("c", 5.0), ("b", 3.0)])
    d3 = d1 + d2
    assert d3["a"] == 1.0
    assert d3["b"] == 5.0
    assert d3["c"] == 5.0
    d3 += {"c": 1.0}
    assert d3["a"] == 1.0
    assert d3["b"] == 5.0
    assert d3["c"] == 6.0
    d4 = {"a": 7.0} + d1
    assert d4["a"] == 8.0
    d5 = d1 + {"a": 9.0}
    assert d5["a"] == 10.0

    d6 = d1 - d2
    assert d6 == {"a": 1.0, "b": -1.0, "c": -5.0}
    d6 -= d3
    assert d6 == {"b": -6.0, "c": -11.0}


def test_ArithmeticDict_add():
    d1 = ArithmeticDict(int, [("a", 1), ("b", 2)])
    d2 = d1 + 3
    assert d2["a"] == 4
    assert d2["b"] == 5
    d2 = 3 + d1
    assert d2["a"] == 4
    assert d2["b"] == 5
    assert d2["c"] == 0
    d3 = d1 + d2
    assert d3["a"] == 5
    assert d3["b"] == 7
    assert d3["c"] == 0


def test_ArithmeticDict_iadd():
    d1 = ArithmeticDict(int, [("a", 1), ("b", 2)])
    d1 += 3
    assert d1["a"] == 4
    assert d1["b"] == 5
    assert d1["c"] == 0


def test_ArithmeticDict_sub():
    d1 = ArithmeticDict(int, [("a", 1), ("b", 2)])
    d2 = d1 - 7
    assert d2["a"] == -6
    assert d2["b"] == -5
    d2 = 3 - d1
    assert d2["a"] == 2
    assert d2["b"] == 1
    assert d2["c"] == 0
    d3 = d1 - d2
    assert d3["a"] == -1
    assert d3["b"] == 1


def test_ArithmeticDict_isub():
    d1 = ArithmeticDict(int, [("a", 1), ("b", 2)])
    d1 -= 7
    assert d1["a"] == -6
    assert d1["b"] == -5
    assert d1["c"] == 0


def test_ArithmeticDict_mul():
    d1 = ArithmeticDict(int, [("a", 1), ("b", 2)])
    d2 = d1 * 3
    assert d2["a"] == 3
    assert d2["b"] == 6
    d2 = 3 * d1
    assert d2["a"] == 3
    assert d2["b"] == 6
    assert d2["c"] == 0
    d3 = d1 * d2
    assert d3["a"] == 3
    assert d3["b"] == 12
    assert d3["c"] == 0


def test_ArithmeticDict_imul():
    d1 = ArithmeticDict(int, [("a", 1), ("b", 2)])
    d1 *= 3
    assert d1["a"] == 3
    assert d1["b"] == 6
    assert d1["c"] == 0


def test_ArithmeticDict_div():
    d1 = ArithmeticDict(int, [("a", 6), ("b", 9)])

    d2 = d1 / 3
    assert d2["a"] == 2
    assert d2["b"] == 3

    d2 = d1.copy() / 3
    assert d2["a"] == 2
    assert d2["b"] == 3

    d2 = 54 / d1
    assert d2["a"] == 9
    assert d2["b"] == 6
    assert d2["c"] == 0

    d1["c"] = 1
    d3 = (6 * d2) / d1
    assert d3["a"] == 9
    assert d3["b"] == 4
    assert d3["c"] == 0


def test_ArithmeticDict_floordiv():
    d1 = ArithmeticDict(int, [("a", 6), ("b", 9)])

    d2 = d1 // 2
    assert d2["a"] == 3
    assert d2["b"] == 4

    d2 = d1.copy() // 5
    assert d2["a"] == 1
    assert d2["b"] == 1

    d2 = 55 // d1
    assert d2["a"] == 9
    assert d2["b"] == 6
    assert d2["c"] == 0

    d1["c"] = 1
    d3 = (6 * d2 + 1) // d1
    assert d3["a"] == 9
    assert d3["b"] == 4
    assert d3["c"] == 1


def test_ArithmeticDict_idiv():
    d1 = ArithmeticDict(int, [("a", 6), ("b", 9)])
    d1 /= 3
    assert d1["a"] == 2
    assert d1["b"] == 3
    assert d1["c"] == 0


def test_ArithmeticDict_div_float():
    d = ArithmeticDict(float)
    d["a"] = 6.0
    d["b"] = 9.0
    t = d / 3.0
    assert t["a"] == 2.0
    assert t["b"] == 3.0


def test_ArithmeticDict_isclose():
    d1 = ArithmeticDict(float)
    d2 = ArithmeticDict(float)
    assert d1.isclose(d2)
    d1["a"] = 2
    assert not d1.isclose(d2)
    d2["a"] = 2 + 1e-15
    assert d1.isclose(d2)
    d2["b"] = 1e-15
    assert not d1.isclose(d2)
    assert d1.isclose(d2, atol=1e-14)


def test_ArithmeticDict_all_non_negative():
    d1 = ArithmeticDict(float)
    assert d1.all_non_negative()
    d1["a"] = 0.1
    assert d1.all_non_negative()
    d1["b"] = 0
    assert d1.all_non_negative()
    d1["b"] -= 1e-15
    assert not d1.all_non_negative()


def test_ArithmeticDict_eq():
    d1 = ArithmeticDict(int, a=1, b=0)
    d2 = ArithmeticDict(int, a=1, b=1)
    d3 = ArithmeticDict(int, a=1)
    assert not d1 == d2
    assert d1 == d3



import re
from .parsing import get_parsing_context
from ..units import fold_constants


class TemplateEvaluator:
    def __init__(self, pattern=r"\${(.*?)}", fmt="${%s}", globals_=None, post_procs=()):
        self.mark = re.compile(pattern)
        self.fmt = fmt
        if globals_ is None:
            globals_ = get_parsing_context()
        self.globals_ = globals_
        self._post_procs = post_procs

    def _post_proc(self, arg):
        for pp in self._post_procs:
            arg = pp(arg)
        return arg

    def __call__(self, template, **kwargs):
        for item in self.mark.findall(template):
            ev = self._post_proc(eval(item, dict(self.globals_, **kwargs)))
            template = template.replace(self.fmt % item, str(ev))
        return template


eval_template = TemplateEvaluator(
    post_procs=(fold_constants, lambda x: str(x).replace(" ", "*"))
)



# -*- coding: utf-8 -*-
"""
General utilities and exceptions.
"""

from collections import defaultdict, namedtuple, OrderedDict

try:
    from collections.abc import ItemsView, Mapping
except ImportError:  # Python 2
    ItemsView = list
    from collections import Mapping
from functools import wraps
from itertools import product
import os
import types
import warnings

from .. import __url__
from .deprecation import Deprecation


def identity(x):
    return x


class NoConvergence(Exception):
    pass


class ChemPyDeprecationWarning(DeprecationWarning):
    pass


def deprecated(*args, **kwargs):
    """Helper to :class:`Deprecation` for using ChemPyDeprecationWarning."""
    return Deprecation(
        *args,
        issues_url=lambda s: __url__ + "/issues/" + s.lstrip("gh-"),
        warning=ChemPyDeprecationWarning,
        **kwargs
    )


warnings.simplefilter(
    os.environ.get("CHEMPY_DEPRECATION_FILTER", "once"), ChemPyDeprecationWarning
)


class DeferredImport(object):
    def __init__(self, modname, arg=None, decorators=None):
        self._modname = modname
        self._arg = arg
        self._decorators = decorators
        self._cache = None

    @property
    def cache(self):
        if self._cache is None:
            if self._arg is None:
                obj = __import__(self._modname)
            else:
                obj = getattr(
                    __import__(self._modname, globals(), locals(), [self._arg]),
                    self._arg,
                )
            if self._decorators is not None:
                for deco in self._decorators:
                    obj = deco(obj)
            self._cache = obj
        return self._cache

    def __getattribute__(self, attr):
        if attr in ("_modname", "_arg", "_cache", "cache", "_decorators"):
            return object.__getattribute__(self, attr)
        else:
            return getattr(self.cache, attr)

    def __call__(self, *args, **kwargs):
        return self.cache(*args, **kwargs)


class NameSpace:
    """Used to wrap, e.g. modules.

    Parameters
    ----------
    default : module
        The underlying module. Acts as a fallback for attribute access.

    Examples
    --------
    >>> import numpy
    >>> my_numpy = NameSpace(numpy)
    >>> my_numpy.array = lambda *args, **kwargs: list(numpy.array(*args, **kwargs))
    >>> isinstance(my_numpy.array([2, 3]), list)
    True
    >>> isinstance(numpy.array([2, 3]), list)
    False

    """

    def __init__(self, default):
        self._NameSpace_default = default
        self._NameSpace_attr_store = {}

    def __getattr__(self, attr):
        if attr.startswith("_NameSpace_"):
            return self.__dict__[attr]
        else:
            try:
                return self._NameSpace_attr_store[attr]
            except KeyError:
                return getattr(self._NameSpace_default, attr)

    def __setattr__(self, attr, val):
        if attr.startswith("_NameSpace_"):
            self.__dict__[attr] = val
        else:
            self._NameSpace_attr_store[attr] = val

    def as_dict(self):
        items = self._NameSpace_default.__dict__.items()
        result = {k: v for k, v in items if not k.startswith("_")}
        result.update(self._NameSpace_attr_store)
        return result


class AttributeContainer(object):
    """Used to turn e.g. a dictionary to a module-like object.

    Parameters
    ----------
    \\*\\*kwargs : dictionary

    Examples
    --------
    >>> def RT(T, const):
    ...     return T*const.molar_gas_constant
    ...
    >>> from quantities import constants
    >>> RT(273.15, constants)
    array(273.15) * R
    >>> my_constants = AttributeContainer(molar_gas_constant=42)
    >>> RT(273.15, my_constants)
    11472.3

    """

    def __init__(self, **kwargs):
        self.__dict__.update(**kwargs)

    def as_dict(self):
        return self.__dict__.copy()

    def __repr__(self):
        return "%s(%s)" % (
            self.__class__.__name__,
            ", ".join(set(dir(self)) - set(dir(object()))),
        )


class AttrDict(dict):
    """Subclass of dict with attribute access to keys"""

    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self


class defaultkeydict(defaultdict):
    """defaultdict where default_factory should have the signature key -> value

    Examples
    --------
    >>> d = defaultkeydict(lambda k: '[%s]' % k, {'a': '[a]', 'b': '[B]'})
    >>> d['a']
    '[a]'
    >>> d['b']
    '[B]'
    >>> d['c']
    '[c]'

    """

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError("Missing key: %s" % key)
        else:
            self[key] = self.default_factory(key)
        return self[key]


def defaultnamedtuple(typename, field_names, defaults=()):
    """Generates a new subclass of tuple with default values.

    Parameters
    ----------
    typename : string
        The name of the class.
    field_names : str or iterable
        An iterable of splitable string.
    defaults : iterable
        Default values for ``field_names``, counting ``[-len(defaults):]``.

    Examples
    --------
    >>> Body = defaultnamedtuple('Body', 'x y z density', (1.0,))
    >>> Body.__doc__
    'Body(x, y, z, density)'
    >>> b = Body(10, z=3, y=5)
    >>> b._asdict() == dict(x=10, y=5, z=3, density=1.0)
    True

    Returns
    -------
    A new tuple subclass named ``typename``

    """
    Tuple = namedtuple(typename, field_names)
    Tuple.__new__.__defaults__ = (None,) * len(Tuple._fields)
    if isinstance(defaults, Mapping):
        Tuple.__new__.__defaults__ = tuple(Tuple(**defaults))
    else:
        nmissing = len(Tuple._fields) - len(defaults)
        defaults = (None,) * nmissing + tuple(defaults)
        Tuple.__new__.__defaults__ = tuple(Tuple(*defaults))
    return Tuple


def multi_indexed_cases(
    od,
    *,
    dict_=OrderedDict,
    apply_keys=None,
    apply_values=None,
    apply_return=list,
    named_index=False
):
    """Returns a list of length-2 tuples

    Each tuple consist of a multi-index (tuple of integers) and a dictionary.

    Parameters
    ----------
    od : OrderedDict
        Maps each key to a number of values. Instances of ``list``, ``tuple``,
        ``types.GeneratorType``, ``collections.abc.ItemsView`` are converted to ``OrderedDict``.
    dict_ : type, optional
        Used in the result (see ``Returns``).
    apply_keys : callable, optional
        Transformation of keys.
    apply_values : callable, optional
        Transformation of values.
    apply_return : callable, optional
        Applied on return value. ``None`` for generator.
    named_index : bool
        Tuple of indices will be a ``namedtuple`` (requires all keys to be ``str``).

    Examples
    --------
    >>> from chempy.util.pyutil import multi_indexed_cases
    >>> cases = multi_indexed_cases([('a', [1, 2, 3]), ('b', [False, True])])
    >>> len(cases)
    6
    >>> midxs, case_kws = zip(*cases)
    >>> midxs[0]
    (0, 0)
    >>> case_kws[0] == {'a': 1, 'b': False}
    True
    >>> d = {'a': 'foo bar'.split(), 'b': 'baz qux'.split()}
    >>> from chempy.util.pyutil import AttrDict
    >>> for nidx, case in multi_indexed_cases(d, dict_=AttrDict, named_index=True):
    ...     if case.a == 'bar' and case.b == 'baz':
    ...         print("{} {}".format(nidx.a, nidx.b))
    ...
    1 0


    Returns
    -------
    List of length-2 tuples, each consisting of one tuple of indices and one dictionary (of type ``dict_``).

    """
    if isinstance(od, OrderedDict):
        pass
    else:
        if hasattr(od, "items"):
            od = od.items()

        if isinstance(od, (list, tuple, types.GeneratorType, ItemsView)):
            od = OrderedDict(od)
        else:
            raise NotImplementedError("Expected an OrderedDict")

    keys, values = tuple(map(apply_keys or identity, od.keys())), tuple(od.values())
    MultiIndex = (
        namedtuple("MultiIndex", keys) if named_index else lambda *args: tuple(args)
    )
    _generator = (
        (
            MultiIndex(*mi),
            dict_(
                [
                    (k, (apply_values or identity)(v[i]))
                    for k, v, i in zip(keys, values, mi)
                ]
            ),
        )
        for mi in product(*map(range, map(len, values)))
    )
    return (apply_return or identity)(_generator)


def memoize(max_nargs=0):
    def decorator(func):
        @wraps(func)
        def wrapper(*args):
            if max_nargs is not None and len(args) > max_nargs:
                raise ValueError("memoization error")
            if args not in wrapper.results:
                wrapper.results[args] = func(*args)
            return wrapper.results[args]

        wrapper.results = {}
        return wrapper

    return decorator



# -*- coding: utf-8 -*-

try:
    import numpy
except ImportError:
    numpy = None
else:
    from ..sulfuric_acid_density_myhre_1998 import (
        sulfuric_acid_density,
        density_from_concentration,
    )

from chempy.util.testing import requires
from chempy.units import units_library


@requires("numpy")
def test_sulfuric_acid_density():
    rho = sulfuric_acid_density(0.1, 298)
    assert abs(1063.8 - rho) < 0.1


@requires("numpy")
def test_density_from_concentration():
    rho = density_from_concentration(1000)
    assert abs(1058.5 - rho) < 0.1


@requires(units_library)
def test_density_from_concentration__units():
    from chempy.units import default_units as units

    rho = density_from_concentration(0.4 * units.molar, units=units)
    assert abs(1.024 * units.kg / units.decimetre ** 3 / rho - 1) < 1e-3



# -*- coding: utf-8 -*-
"""
Convenience functions for presenting reaction systems in tables.
"""


import os
import shutil
import subprocess
import tempfile

from ..printing import latex
from ..kinetics.rates import RadiolyticBase
from ..units import to_unitless, get_derived_unit

tex_templates = {
    "document": {
        "default": r"""
\documentclass[a4paper,9pt]{article}
\pagestyle{empty}
\usepackage[paper=a4paper,margin=1cm]{geometry}
%(usepkg)s
\hypersetup{
  bookmarksnumbered=true,
  breaklinks=false,
  raiselinks=true,
  pdfborder={0 0 0},
  colorlinks=true,
  plainpages=false,
  pdfstartview={FitH},
  pdfcreator={LaTeX with hyperref package},
  citecolor=teal,
  linkcolor=red,
  urlcolor=blue,
}
\begin{document}
%(begins)s
%(table)s
%(ends)s
\end{document}
"""
    },
    "table": {
        "default": r"""
\begin{%(table_env)s}
\centering
\label{tab:%(label)s}
\caption[%(short_cap)s]{%(long_cap)s}
\begin{tabular}{%(alignment)s}
\toprule
%(header)s
\midrule
%(body)s
\bottomrule
\end{tabular}
\end{%(table_env)s}""",
        "longtable": r"""
\begin{%(table_env)s}{%(alignment)s}
\caption[%(short_cap)s]{%(long_cap)s
\label{tab:%(label)s}}\\
\toprule
%(header)s
\midrule
%(body)s
\bottomrule
\end{%(table_env)s}""",
    },
}


def render_tex_to_pdf(contents, texfname, pdffname, output_dir, save):
    """Generates a pdf from a tex file by calling pdflatex

    Parameters
    ----------
    contents : str
    texfname : path
    pdffname : path
    output_dir : path
    save : path or bool or str(bool)

    """
    created_tempdir = False
    try:
        if output_dir is None:
            output_dir = tempfile.mkdtemp()
            created_tempdir = True
        texpath = os.path.join(output_dir, texfname)
        pdfpath = os.path.join(output_dir, pdffname)
        cmds = ["pdflatex", "-halt-on-error", "-interaction", "batchmode", texfname]
        with open(texpath, "wt") as ofh:
            ofh.write(contents)
            ofh.flush()
        with open(pdfpath + ".out", "wb") as logfile:
            p = subprocess.Popen(cmds, cwd=output_dir, stdout=logfile, stderr=logfile)
            retcode = p.wait()
            p = subprocess.Popen(cmds, cwd=output_dir, stdout=logfile, stderr=logfile)
            retcode += p.wait()
        if retcode:
            fmtstr = "{}\n returned with exit status {}"
            raise RuntimeError(fmtstr.format(" ".join(cmds), retcode))
        else:
            return pdfpath
    finally:
        if save is True or save == "True":
            pass
        else:
            if save is False or save == "False":
                if created_tempdir:
                    shutil.rmtree(output_dir)
            else:
                # interpret path to copy pdf to.
                if not os.path.samefile(pdfpath, save):
                    shutil.copy(pdfpath, save)


def rsys2tablines(
    rsys,
    rref0=1,
    coldelim=" & ",
    tex=True,
    ref_fmt=None,
    unit_registry=None,
    unit_fmt="{}",
    k_fmt="%.4g",
):
    """
    Generates a table representation of a ReactionSystem.

    Parameters
    ----------
    rsys : ReactionSystem
    rref0 : integer
        default start of index counter (default: 1)
    coldelim : string
        column delimiter (default: ' & ')
    tex : bool
        use latex formatted output (default: True)
    ref_fmt : string or callable
        format string of ``ref`` attribute of reactions
    unit_registry : unit registry
        optional (default: None)
    """
    if ref_fmt is None:

        def _doi(s):
            return r"\texttt{\href{http://dx.doi.org/" + s + "}{doi:" + s + "}}"

        def ref_fmt(s):
            if s is None:
                return "None"
            if tex:
                if isinstance(s, dict):
                    return _doi(s["doi"])
                if s.startswith("doi:"):
                    return _doi(s[4:])
            return s

    def _wrap(s):
        if tex:
            return "\\ensuremath{" + s + "}"
        else:
            return s

    lines = []
    for ri, rxn in enumerate(rsys.rxns):
        rxn_ref = rxn.ref
        if isinstance(rxn.param, RadiolyticBase):
            if unit_registry is not None:
                kunit = get_derived_unit(unit_registry, "radiolytic_yield")
                k = k_fmt % to_unitless(rxn.param.args[0], kunit)
                k_unit_str = (
                    kunit.dimensionality.latex.strip("$")
                    if tex
                    else kunit.dimensionality
                )
        else:
            if unit_registry is not None:
                kunit = get_derived_unit(unit_registry, "concentration") ** (
                    1 - rxn.order()
                ) / get_derived_unit(unit_registry, "time")
                try:
                    k = k_fmt % to_unitless(rxn.param, kunit)
                    k_unit_str = (
                        kunit.dimensionality.latex.strip("$")
                        if tex
                        else kunit.dimensionality
                    )
                except Exception:
                    k, k_unit_str = rxn.param.equation_as_string(k_fmt, tex)
            else:
                k_unit_str = "-"
                if isinstance(k_fmt, str):
                    k = k_fmt % rxn.param
                else:
                    k = k_fmt(rxn.param)
        latex_kw = dict(with_param=False, with_name=False)
        if tex:
            latex_kw["substances"] = rsys.substances
            latex_kw["Reaction_around_arrow"] = (
                "}}" + coldelim + "\\ensuremath{{",
                "}}" + coldelim + "\\ensuremath{{",
            )
        else:
            latex_kw["Reaction_around_arrow"] = (coldelim,) * 2
            latex_kw["Reaction_arrow"] = "->"
        lines.append(
            coldelim.join(
                [
                    str(rref0 + ri),
                    ("\\ensuremath{%s}" if tex else "%s") % latex(rxn, **latex_kw),
                    _wrap(k),
                    unit_fmt.format(_wrap(k_unit_str)),
                    ref_fmt(rxn_ref) if callable(ref_fmt) else ref_fmt.format(rxn_ref),
                ]
            )
        )

    return lines


def rsys2table(
    rsys,
    table_template=None,
    table_template_dict=None,
    param_name="Rate constant",
    **kwargs
):
    r"""
    Renders user provided table_template with table_template_dict which
    also has 'body' entry generated from `rsys2tablines`.

    Defaults is LaTeX table requiring booktabs package to be used
    (add \usepackage{booktabs} to preamble).

    Parameters
    ----------
    rsys : ReactionSystem
    table_template : string
    table_tempalte_dict : dict used to render table_template (excl. "body")
    param_name : str
        Column header for parameter column
    longtable : bool
        use longtable in defaults. (default: False)
    **kwargs :
        passed onto rsys2tablines

    """
    siunitx = kwargs.pop("siunitx", False)
    line_term = r" \\"
    defaults = {
        "table_env": "longtable" if kwargs.pop("longtable", False) else "table",
        "alignment": "llllSll" if siunitx else "lllllll",
        "header": kwargs.get("coldelim", " & ").join(
            ["Id.", "Reactants", "", "Products", "{%s}" % param_name, "Unit", "Ref"]
        )
        + line_term,
        "short_cap": rsys.name,
        "long_cap": rsys.name,
        "label": (rsys.name or "None").lower(),
    }

    if table_template_dict is None:
        table_template_dict = defaults
    else:
        for k, v in defaults:
            if k not in table_template_dict:
                table_template_dict[k] = v

    if "body" in table_template_dict:
        raise KeyError("There is already a 'body' key in table_template_dict")
    if "k_fmt" not in kwargs:
        kwargs["k_fmt"] = r"\num{%.4g}" if siunitx else "%.4g"
    table_template_dict["body"] = (line_term + "\n").join(
        rsys2tablines(rsys, **kwargs)
    ) + line_term

    if table_template is None:
        if table_template_dict["table_env"] == "longtable":
            table_template = tex_templates["table"]["longtable"]
        else:
            table_template = tex_templates["table"]["default"]

    return table_template % table_template_dict


def rsys2pdf_table(
    rsys,
    output_dir=None,
    doc_template=None,
    doc_template_dict=None,
    save=True,
    landscape=False,
    **kwargs
):
    """
    Convenience function to render a ReactionSystem as
    e.g. a pdf using e.g. pdflatex.

    Parameters
    ----------
    rsys : ReactionSystem
    output_dir : path to output directory
        (default: system's temporary folder)
    doc_template : string
        LaTeX boiler plate temlpate including preamble,
        document environment etc.
    doc_template_dict : dict (string -> string)
        dict used to render temlpate (excl. 'table')
    longtable : bool
        use longtable in defaults. (default: False)
    **kwargs :
        passed on to `rsys2table`
    """
    if doc_template is None:
        doc_template = tex_templates["document"]["default"]
    lscape = ["pdflscape" if landscape == "pdf" else "lscape"] if landscape else []
    _pkgs = [
        "booktabs",
        "amsmath",
        ("pdftex,colorlinks,unicode=True", "hyperref"),
    ] + lscape
    if kwargs.get("longtable", False):
        _pkgs += ["longtable"]
    if kwargs.get("siunitx", False):
        _pkgs += ["siunitx"]
    _envs = ["tiny"] + (["landscape"] if landscape else [])
    defaults = {
        "usepkg": "\n".join(
            [
                (r"\usepackage" + ("[%s]" if isinstance(pkg, tuple) else "") + "{%s}")
                % pkg
                for pkg in _pkgs
            ]
        ),
        "begins": "\n".join([r"\begin{%s}" % env for env in _envs]),
        "ends": "\n".join([r"\end{%s}" % env for env in _envs[::-1]]),
    }

    if doc_template_dict is None:
        doc_template_dict = defaults
    else:
        for k, v in defaults:
            if k not in doc_template_dict:
                doc_template_dict[k] = v

    if "table" in doc_template_dict:
        raise KeyError("There is already a 'table' key in doc_template_dict")
    doc_template_dict["table"] = rsys2table(rsys, **kwargs)

    contents = doc_template % doc_template_dict

    if isinstance(save, str) and save.endswith(".pdf"):
        texfname = save.rstrip(".pdf") + ".tex"
        pdffname = save
    else:
        texfname = "output.tex"
        pdffname = "output.pdf"
    return render_tex_to_pdf(contents, texfname, pdffname, output_dir, save)



# -*- coding: utf-8 -*-

from ..nernst import nernst_potential
from chempy.util.testing import requires
from chempy.units import default_units, default_constants, units_library, allclose


def test_nernst_potential():
    """
    Test cases obtained from textbook examples of Nernst potential in cellular
    membranes. 310K = 37C, typical mammalian cell environment temperature.
    """
    # Sodium in cells
    assert abs(1000 * nernst_potential(145, 15, 1, 310) - 60.605) < 1e-4
    # Potassium in cells
    assert abs(1000 * nernst_potential(4, 150, 1, 310) - (-96.8196)) < 1e-4
    # Calcium in cells
    assert abs(1000 * nernst_potential(2, 7e-5, 2, 310) - 137.0436) < 1e-4
    # Chloride in cells
    assert abs(1000 * nernst_potential(110, 10, -1, 310) - (-64.0567)) < 1e-4


@requires(units_library)
def test_nernst_potential__units():
    J = default_units.joule
    K = default_units.kelvin
    coulomb = default_units.coulomb
    v = nernst_potential(145, 15, 1, 310 * K, default_constants)
    assert allclose(1000 * v, 60.605 * J / coulomb, rtol=1e-4)



# -*- coding: utf-8 -*-

import pytest

from .._solution import Solution, QuantityDict
from ..util.testing import requires
from ..units import magnitude, units_library, to_unitless, default_units as u


@requires(units_library)
def test_QuantityDict():
    # QuantityDict * scalar_quantity
    c = QuantityDict(u.molar, {})
    c["H2O"] = 55.4 * u.molar
    with pytest.raises(ValueError):
        c["HCl"] = 3 * u.kg

    with pytest.raises(ValueError):
        QuantityDict(u.molar, {"a": u.mole})

    V = 0.4 * u.dm3
    n = c * V
    assert isinstance(n, QuantityDict)

    # For the following to work major changes to quantities needs to be made:
    # n = V*c
    # assert isinstance(n, QuantityDict)

    assert n.isclose({"H2O": 55.4 * 0.4 * u.mol})
    assert abs(to_unitless(n["H2O"], u.mol) - 55.4 * 0.4) < 1e-14

    c2 = c.rescale(u.mol / u.cm3)
    assert abs(magnitude(c2["H2O"]) - 55.4e-3) < 1e-6


def _get_s1_s2():
    s1 = Solution(0.1 * u.dm3, {"CH3OH": 0.1 * u.molar})
    s2 = Solution(
        0.3 * u.dm3,
        {"CH3OH": 0.4 * u.molar, "Na+": 2e-3 * u.molar, "Cl-": 2e-3 * u.molar},
    )
    return s1, s2


@requires(units_library)
def test_Solution__add():
    s1, s2 = _get_s1_s2()
    s3 = s1 + s2
    assert abs(to_unitless(s3.volume - 4e-4 * u.m ** 3, u.dm3)) < 1e-15


@requires(units_library)
def test_Solution__isclose():
    s1, s2 = _get_s1_s2()
    s3 = s1 + s2
    assert s3.concentrations.isclose(
        {"CH3OH": 0.325 * u.molar, "Na+": 1.5e-3 * u.molar, "Cl-": 1.5e-3 * u.molar}
    )


@requires(units_library)
def test_Solution__dissolve():
    s1, s2 = _get_s1_s2()
    s4 = (s1 + s2).dissolve({"CH3OH": 1 * u.gram})
    mw = 12.011 + 4 * 1.008 + 15.999
    assert abs(s4.concentrations["CH3OH"] - (0.325 + 1 / mw / 0.4) * u.molar) < 1e-7


@requires(units_library)
def test_Solution__withdraw():
    s1, s2 = _get_s1_s2()
    s3 = s1 + s2
    s4 = s3.withdraw(0.2 * u.dm3)
    assert s4 == s3
    assert s4 == (s1 + s2).withdraw(0.2 * u.dm3)
    assert s4 != s1 + s2
